"""
TestConstructionFactory test and validate the city model structure construction parameters
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from unittest import TestCase

from hub.imports.geometry_factory import GeometryFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.helpers.dictionaries import Dictionaries


class TestConstructionFactory(TestCase):
  """
  TestConstructionFactory TestCase
  """
  def setUp(self) -> None:
    """
    Configure test environment
    :return:
    """
    self._city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()

  def _get_citygml(self, file):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('citygml', path=file_path).city
    self.assertIsNotNone(self._city, 'city is none')
    self.assertIsNotNone(self._city.level_of_detail.geometry, 'wrong construction level of detail')
    return self._city

  @staticmethod
  def _internal_function(function_format, original_function):
    if function_format == 'hft':
      new_function = Dictionaries().hft_function_to_hub_function[original_function]
    elif function_format == 'pluto':
      new_function = Dictionaries().pluto_function_to_hub_function[original_function]
    else:
      raise Exception('Function key not recognized. Implemented only "hft" and "pluto"')
    return new_function

  def test_citygml_function(self):
    """
    Test city objects' functions in the city
    """
    # case 1: hft
    file = 'one_building_in_kelowna.gml'
    function_format = 'hft'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.function = self._internal_function(function_format, building.function)
      self.assertEqual('residential', building.function, 'format hft')

    # case 2: Pluto
    file = 'pluto_building.gml'
    function_format = 'pluto'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.function = self._internal_function(function_format, building.function)
      self.assertEqual('education', building.function, 'format pluto')

    # case 3: Alkis
    file = 'one_building_in_kelowna_alkis.gml'
    function_format = 'alkis'
    city = self._get_citygml(file)
    for building in city.buildings:
      self.assertRaises(Exception, lambda: self._internal_function(function_format, building.function))

  def _check_buildings(self, city):
    for building in city.buildings:
      self.assertIsNotNone(building.name, 'building name is none')
      self.assertIsNotNone(building.type, 'building type is none')
      self.assertIsNotNone(building.volume, 'building volume is none')
      self.assertIsNotNone(building.detailed_polyhedron, 'building detailed polyhedron is none')
      self.assertIsNotNone(building.simplified_polyhedron, 'building simplified polyhedron is none')
      self.assertIsNotNone(building.surfaces, 'building surfaces is none')
      self.assertIsNotNone(building.centroid, 'building centroid is none')
      self.assertIsNotNone(building.max_height, 'building max_height is none')
      self.assertEqual(len(building.external_temperature), 0, 'building external temperature is calculated')
      self.assertEqual(len(building.global_horizontal), 0, 'building global horizontal is calculated')
      self.assertEqual(len(building.diffuse), 0, 'building diffuse is calculated')
      self.assertEqual(len(building.beam), 0, 'building beam is calculated')
      self.assertIsNotNone(building.lower_corner, 'building lower corner is none')
      self.assertEqual(len(building.sensors), 0, 'building sensors are assigned')
      self.assertIsNotNone(building.internal_zones, 'no internal zones created')
      self.assertIsNotNone(building.grounds, 'building grounds is none')
      self.assertIsNotNone(building.walls, 'building walls is none')
      self.assertIsNotNone(building.roofs, 'building roofs is none')
      self.assertIsNotNone(building.internal_walls, 'building internal walls is none')
      self.assertIsNone(building.basement_heated, 'building basement_heated is not none')
      self.assertIsNone(building.attic_heated, 'building attic_heated is not none')
      self.assertIsNone(building.terrains, 'building terrains is not none')
      self.assertIsNotNone(building.year_of_construction, 'building year_of_construction is none')
      self.assertIsNotNone(building.function, 'building function is none')
      self.assertIsNotNone(building.average_storey_height, 'building average_storey_height is none')
      self.assertIsNotNone(building.storeys_above_ground, 'building storeys_above_ground is none')
      self.assertEqual(len(building.heating_demand), 0, 'building heating is not none')
      self.assertEqual(len(building.cooling_demand), 0, 'building cooling is not none')
      self.assertIsNotNone(building.eave_height, 'building eave height is none')
      self.assertIsNotNone(building.roof_type, 'building roof type is none')
      self.assertIsNotNone(building.floor_area, 'building floor_area is none')
      self.assertIsNone(building.households, 'building households is not none')
      self.assertFalse(building.is_conditioned, 'building is conditioned')
      self.assertIsNotNone(building.shell, 'building shell is none')

  def _check_thermal_zones(self, internal_zone):
    for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
      self.assertIsNotNone(thermal_zone.id, 'thermal_zone id is none')
      self.assertIsNotNone(thermal_zone.footprint_area, 'thermal_zone floor area is none')
      self.assertTrue(len(thermal_zone.thermal_boundaries) > 0, 'thermal_zone thermal_boundaries not defined')
      self.assertIsNotNone(thermal_zone.additional_thermal_bridge_u_value, 'additional_thermal_bridge_u_value is none')
      self.assertIsNotNone(thermal_zone.effective_thermal_capacity, 'thermal_zone effective_thermal_capacity is none')
      self.assertIsNotNone(thermal_zone.infiltration_rate_system_off,
                           'thermal_zone infiltration_rate_system_off is none')
      self.assertIsNotNone(thermal_zone.infiltration_rate_system_on, 'thermal_zone infiltration_rate_system_on is none')
      self.assertIsNotNone(thermal_zone.volume, 'thermal_zone volume is none')
      self.assertIsNone(thermal_zone.ordinate_number, 'thermal_zone ordinate number is not none')
      self.assertIsNotNone(thermal_zone.view_factors_matrix, 'thermal_zone view factors matrix is none')
      self.assertIsNotNone(thermal_zone.total_floor_area, 'thermal zone total_floor_area is none')
      self.assertIsNone(thermal_zone.usage_name, 'thermal_zone usage is not none')
      self.assertIsNone(thermal_zone.hours_day, 'thermal_zone hours a day is not none')
      self.assertIsNone(thermal_zone.days_year, 'thermal_zone days a year is not none')
      self.assertIsNone(thermal_zone.mechanical_air_change, 'thermal_zone mechanical air change is not none')
      self.assertIsNone(thermal_zone.occupancy, 'thermal_zone occupancy is not none')
      self.assertIsNone(thermal_zone.lighting, 'thermal_zone lighting is not none')
      self.assertIsNone(thermal_zone.appliances, 'thermal_zone appliances is not none')
      self.assertIsNone(thermal_zone.thermal_control, 'thermal_zone thermal control is not none')
      self.assertIsNone(thermal_zone.internal_gains, 'thermal_zone internal gains not returns none')

  def _check_thermal_boundaries(self, thermal_zone):
      for thermal_boundary in thermal_zone.thermal_boundaries:
        self.assertIsNotNone(thermal_boundary.id, 'thermal_boundary id is none')
        self.assertIsNotNone(thermal_boundary.parent_surface, 'thermal_boundary surface is none')
        self.assertIsNotNone(thermal_boundary.thermal_zones, 'thermal_boundary delimits no thermal zone')
        self.assertIsNotNone(thermal_boundary.opaque_area, 'thermal_boundary area is none')
        self.assertIsNotNone(thermal_boundary.thickness, 'thermal_boundary thickness is none')
        self.assertIsNotNone(thermal_boundary.type, 'thermal_boundary type is none')
        self.assertIsNotNone(thermal_boundary.thermal_openings, 'thermal_openings is none')
        self.assertIsNotNone(thermal_boundary.window_ratio, 'window_ratio is none')
        self.assertIsNone(thermal_boundary.windows_areas, 'windows_areas is not none')
        self.assertIsNotNone(thermal_boundary.u_value, 'u_value is none')
        self.assertIsNotNone(thermal_boundary.hi, 'hi is none')
        self.assertIsNotNone(thermal_boundary.he, 'he is none')
        self.assertIsNotNone(thermal_boundary.internal_surface, 'virtual_internal_surface is none')
        self.assertIsNotNone(thermal_boundary.layers, 'layers is not none')

  def _check_thermal_openings(self, thermal_boundary):
    for thermal_opening in thermal_boundary.thermal_openings:
      self.assertIsNotNone(thermal_opening.id, 'thermal opening id is not none')
      self.assertIsNotNone(thermal_opening.area, 'thermal opening area is not none')
      self.assertIsNotNone(thermal_opening.frame_ratio, 'thermal opening frame_ratio is none')
      self.assertIsNotNone(thermal_opening.g_value, 'thermal opening g_value is none')
      self.assertIsNotNone(thermal_opening.overall_u_value, 'thermal opening overall_u_value is none')
      self.assertIsNotNone(thermal_opening.hi, 'thermal opening hi is none')
      self.assertIsNotNone(thermal_opening.he, 'thermal opening he is none')

  def _check_surfaces(self, thermal_boundary):
    external_surface = thermal_boundary.external_surface
    internal_surface = thermal_boundary.internal_surface
    self.assertIsNotNone(external_surface.short_wave_reflectance,
                         'external surface short_wave_reflectance id is not none')
    self.assertIsNotNone(external_surface.long_wave_emittance, 'external surface long_wave_emittance id is not none')
    self.assertIsNotNone(internal_surface.short_wave_reflectance,
                         'external surface short_wave_reflectance id is not none')
    self.assertIsNotNone(internal_surface.long_wave_emittance, 'external surface long_wave_emittance id is not none')

  def test_city_with_construction_extended_library(self):
    """
    Enrich the city with the construction information and verify it
    """
    file = 'one_building_in_kelowna.gml'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.year_of_construction = 1980
      building.function = self._internal_function('hft', building.function)
    ConstructionFactory('nrcan', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

    file = 'pluto_building.gml'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.year_of_construction = 1980
      building.function = self._internal_function('pluto', building.function)
    ConstructionFactory('nrcan', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

    file = 'pluto_building.gml'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.year_of_construction = 2006
      building.function = self._internal_function('pluto', building.function)
    ConstructionFactory('nrel', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

    file = 'one_building_in_kelowna.gml'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.year_of_construction = 1980
      building.function = self._internal_function('hft', building.function)
    ConstructionFactory('nrcan', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

    file_path = (self._example_path / 'test.geojson').resolve()
    self._city = GeometryFactory('geojson',
                                 path=file_path,
                                 height_field='citygml_me',
                                 year_of_construction_field='ANNEE_CONS',
                                 function_field='CODE_UTILI',
                                 function_to_hub=Dictionaries().montreal_function_to_hub_function).city

    ConstructionFactory('nrcan', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

  def test_nrcan_construction_factory(self):
    file = 'test.geojson'
    file_path = (self._example_path / file).resolve()
    city = GeometryFactory('geojson',
                           path=file_path,
                           height_field='citygml_me',
                           year_of_construction_field='ANNEE_CONS',
                           function_field='CODE_UTILI',
                           function_to_hub=Dictionaries().montreal_function_to_hub_function).city
    ConstructionFactory('nrcan', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

  def test_eilat_construction_factory(self):
    file = 'eilat.geojson'
    file_path = (self._example_path / file).resolve()
    city = GeometryFactory('geojson',
                           path=file_path,
                           height_field='heightmax',
                           year_of_construction_field='ANNEE_CONS',
                           function_field='CODE_UTILI',
                           function_to_hub=Dictionaries().eilat_function_to_hub_function).city
    ConstructionFactory('eilat', city).enrich()

    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)

  def test_palma_construction_factory(self):
    file = 'palma_test_file.geojson'
    file_path = (self._example_path / file).resolve()
    city = GeometryFactory(file_type='geojson',
                           path=file_path,
                           height_field='measuredHeight',
                           year_of_construction_field='yearOfConstruction',
                           function_field='usage',
                           function_to_hub=Dictionaries().palma_function_to_hub_function).city
    ConstructionFactory('palma', city).enrich()
    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self._check_thermal_zones(internal_zone)
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_boundaries(thermal_zone)
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.layers, 'layers is none')
            self._check_thermal_openings(thermal_boundary)
            self._check_surfaces(thermal_boundary)


"""
Geometry helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import math
from pathlib import Path
from typing import Dict

from PIL import Image
from trimesh import Trimesh
from trimesh import intersections
import numpy as np

from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.attributes.polyhedron import Polyhedron
from hub.helpers.location import Location


class MapPoint:
  """
  Map point class
  """
  def __init__(self, x, y):
    self._x = int(x)
    self._y = int(y)

  @property
  def x(self):
    """
    Get X Coordinate
    """
    return self._x

  @property
  def y(self):
    """
    Get Y Coordinate
    """
    return self._y

  def __str__(self):
    return f'({self.x}, {self.y})'

  def __len__(self):
    return 1

  def __getitem__(self, index):
    if index == 0:
      return self._x
    if index == 1:
      return self._y
    raise IndexError('Index error')


class GeometryHelper:
  """
  Geometry helper class
  """
  # todo: complete dictionary
  srs_transformations = {
    'urn:adv:crs:ETRS89_UTM32*DE_DHHN92_NH': 'epsg:25832'
  }

  @staticmethod
  def factor():
    """
    Set minimap resolution
    :return: None
    """
    return 0.5

  def __init__(self, delta=0, area_delta=0):
    self._delta = delta
    self._area_delta = area_delta

  @staticmethod
  def coordinate_to_map_point(coordinate, city):
    """
    Transform a real world coordinate to a minimap one
    :param coordinate: real world coordinate
    :param city: current city
    :return: None
    """
    factor = GeometryHelper.factor()
    return MapPoint(
      ((coordinate[0] - city.lower_corner[0]) * factor), ((coordinate[1] - city.lower_corner[1]) * factor)
    )

  @staticmethod
  def city_mapping(city, building_names=None, plot=False) -> Dict:
    """
    :param city: city to be mapped
    :param building_names: list of building names to be mapped or None
    :param plot: True if minimap image should be displayed
    :return: shared_information dictionary
    """
    lines_information = {}
    if building_names is None:
      building_names = [b.name for b in city.buildings]
    factor = GeometryHelper.factor()
    x = math.ceil((city.upper_corner[0] - city.lower_corner[0]) * factor) + 1
    y = math.ceil((city.upper_corner[1] - city.lower_corner[1]) * factor) + 1
    city_map = [['' for _ in range(y + 1)] for _ in range(x + 1)]
    map_info = [[{} for _ in range(y + 1)] for _ in range(x + 1)]
    img = Image.new('RGB', (x + 1, y + 1), "black")  # create a new black image
    city_image = img.load()  # create the pixel map
    for building_name in building_names:
      building = city.city_object(building_name)
      line = 0
      for ground in building.grounds:
        length = len(ground.perimeter_polygon.coordinates) - 1
        for i, coordinate in enumerate(ground.perimeter_polygon.coordinates):

          j = i + 1
          if i == length:
            j = 0
          next_coordinate = ground.perimeter_polygon.coordinates[j]
          distance = GeometryHelper.distance_between_points(coordinate, next_coordinate)
          steps = int(distance * factor * 2)
          if steps == 0:
            continue
          delta_x = (next_coordinate[0] - coordinate[0]) / steps
          delta_y = (next_coordinate[1] - coordinate[1]) / steps

          for k in range(0, steps):
            new_coordinate = (coordinate[0] + (delta_x * k), coordinate[1] + (delta_y * k))
            point = GeometryHelper.coordinate_to_map_point(new_coordinate, city)
            x = point.x
            y = point.y
            if city_map[x][y] == '':
              city_map[x][y] = building.name
              map_info[x][y] = {
                'line_start': (coordinate[0], coordinate[1]),
                'line_end': (next_coordinate[0], next_coordinate[1]),
              }
              city_image[x, y] = (100, 0, 0)
            elif city_map[x][y] != building.name:
              neighbour = city.city_object(city_map[x][y])
              neighbour_info = map_info[x][y]

              # prepare the keys
              neighbour_start_coordinate = f'{GeometryHelper.coordinate_to_map_point(neighbour_info["line_start"], city)}'
              building_start_coordinate = f'{GeometryHelper.coordinate_to_map_point(coordinate, city)}'
              neighbour_key = f'{neighbour.name}_{neighbour_start_coordinate}_{building_start_coordinate}'
              building_key = f'{building.name}_{building_start_coordinate}_{neighbour_start_coordinate}'

              # Add my neighbour info to my shared lines
              if building.name in lines_information and neighbour_key in lines_information[building.name]:
                shared_points = int(lines_information[building.name][neighbour_key]['shared_points'])
                lines_information[building.name][neighbour_key]['shared_points'] = shared_points + 1
              else:
                if building.name not in lines_information:
                  lines_information[building.name] = {}
                lines_information[building.name][neighbour_key] = {
                  'neighbour_name': neighbour.name,
                  'line_start': (coordinate[0], coordinate[1]),
                  'line_end': (next_coordinate[0], next_coordinate[1]),
                  'neighbour_line_start': neighbour_info['line_start'],
                  'neighbour_line_end': neighbour_info['line_end'],
                  'coordinate_start': f"{GeometryHelper.coordinate_to_map_point(coordinate, city)}",
                  'coordinate_end': f"{GeometryHelper.coordinate_to_map_point(next_coordinate, city)}",
                  'neighbour_start': f"{GeometryHelper.coordinate_to_map_point(neighbour_info['line_start'], city)}",
                  'neighbour_end': f"{GeometryHelper.coordinate_to_map_point(neighbour_info['line_end'], city)}",
                  'shared_points': 1
                }

              # Add my info to my neighbour shared lines
              if neighbour.name in lines_information and building_key in lines_information[neighbour.name]:
                shared_points = int(lines_information[neighbour.name][building_key]['shared_points'])
                lines_information[neighbour.name][building_key]['shared_points'] = shared_points + 1
              else:
                if neighbour.name not in lines_information:
                  lines_information[neighbour.name] = {}
                lines_information[neighbour.name][building_key] = {
                  'neighbour_name': building.name,
                  'line_start': neighbour_info['line_start'],
                  'line_end': neighbour_info['line_end'],
                  'neighbour_line_start': (coordinate[0], coordinate[1]),
                  'neighbour_line_end': (next_coordinate[0], next_coordinate[1]),
                  'neighbour_start': f"{GeometryHelper.coordinate_to_map_point(coordinate, city)}",
                  'neighbour_end': f"{GeometryHelper.coordinate_to_map_point(next_coordinate, city)}",
                  'coordinate_start': f"{GeometryHelper.coordinate_to_map_point(neighbour_info['line_start'], city)}",
                  'coordinate_end': f"{GeometryHelper.coordinate_to_map_point(neighbour_info['line_end'], city)}",
                  'shared_points': 1
                }

              if building.neighbours is None:
                building.neighbours = [neighbour]
              elif neighbour not in building.neighbours:
                building.neighbours.append(neighbour)
              if neighbour.neighbours is None:
                neighbour.neighbours = [building]
              elif building not in neighbour.neighbours:
                neighbour.neighbours.append(building)
          line += 1

    if plot:
      img.show()
    return lines_information

  @staticmethod
  def segment_list_to_trimesh(lines) -> Trimesh:
    """
    :param lines: lines
    :return: Transform a list of segments into a Trimesh
    """
    # todo: trimesh has a method for this
    line_points = [lines[0][0], lines[0][1]]
    lines.remove(lines[0])
    while len(lines) > 1:
      i = 0
      for line in lines:
        i += 1
        if GeometryHelper.distance_between_points(line[0], line_points[len(line_points) - 1]) < 1e-8:
          line_points.append(line[1])
          lines.pop(i - 1)
          break
        if GeometryHelper.distance_between_points(line[1], line_points[len(line_points) - 1]) < 1e-8:
          line_points.append(line[0])
          lines.pop(i - 1)
          break
    polyhedron = Polyhedron(Polygon(line_points).triangles)
    trimesh = Trimesh(polyhedron.vertices, polyhedron.faces)
    return trimesh

  @staticmethod
  def _merge_meshes(mesh1, mesh2):
    v_1 = mesh1.vertices
    f_1 = mesh1.faces
    v_2 = mesh2.vertices
    f_2 = mesh2.faces
    length = len(v_1)
    v_merge = np.concatenate((v_1, v_2))
    f_merge = np.asarray(f_1)

    for item in f_2:
      point1 = item.item(0) + length
      point2 = item.item(1) + length
      point3 = item.item(2) + length
      surface = np.asarray([point1, point2, point3])
      f_merge = np.concatenate((f_merge, [surface]))

    mesh_merge = Trimesh(vertices=v_merge, faces=f_merge)
    mesh_merge.fix_normals()

    return mesh_merge

  @staticmethod
  def divide_mesh_by_plane(trimesh, normal_plane, point_plane):
    """
    Divide a mesh by a plane
    :param trimesh: Trimesh
    :param normal_plane: [x, y, z]
    :param point_plane: [x, y, z]
    :return: [Trimesh]
    """
    # The first mesh returns the positive side of the plane and the second the negative side.
    # If the plane does not divide the mesh (i.e. it does not touch it, or it is coplanar with one or more faces),
    # then it returns only the original mesh.
    # todo: review split method in https://github.com/mikedh/trimesh/issues/235,
    #  once triangulate_polygon in Polygon class is solved

    normal_plane_opp = [None] * len(normal_plane)
    for index, normal in enumerate(normal_plane):
      normal_plane_opp[index] = - normal

    section_1 = intersections.slice_mesh_plane(trimesh, normal_plane, point_plane)
    if section_1 is None:
      return [trimesh]
    lines = list(intersections.mesh_plane(trimesh, normal_plane, point_plane))
    cap = GeometryHelper.segment_list_to_trimesh(lines)
    trimesh_1 = GeometryHelper._merge_meshes(section_1, cap)

    section_2 = intersections.slice_mesh_plane(trimesh, normal_plane_opp, point_plane)
    if section_2 is None:
      return [trimesh_1]
    trimesh_2 = GeometryHelper._merge_meshes(section_2, cap)

    return [trimesh_1, trimesh_2]

  @staticmethod
  def get_location(latitude, longitude) -> Location:
    """
    Get Location from latitude and longitude
    :param latitude: Latitude
    :param longitude: Longitude
    :return: Location
    """
    _data_path = Path(Path(__file__).parent.parent / 'data/geolocation/cities15000.txt').resolve()
    latitude = float(latitude)
    longitude = float(longitude)
    distance = math.inf
    country = 'Unknown'
    city = 'Unknown'
    region_code = 'Unknown'
    with open(_data_path, 'r', encoding='utf-8') as file:
      for _, line in enumerate(file):
        fields = line.split('\t')
        file_city_name = fields[2]
        file_latitude = float(fields[4])
        file_longitude = float(fields[5])
        file_country_code = fields[8]
        admin1_code = fields[10]
        admin2_code = fields[11]

        new_distance = math.sqrt(pow((latitude - file_latitude), 2) + pow((longitude - file_longitude), 2))
        if distance > new_distance:
          distance = new_distance
          country = file_country_code
          city = file_city_name
          region_code = f'{file_country_code}.{admin1_code}.{admin2_code}'
    return Location(country, city, region_code, latitude, longitude)

  @staticmethod
  def distance_between_points(vertex1, vertex2):
    """
    distance between points in an n-D Euclidean space
    :param vertex1: point or vertex
    :param vertex2: point or vertex
    :return: float
    """
    power = 0
    for dimension, current_vertex in enumerate(vertex1):
      power += math.pow(vertex2[dimension] - current_vertex, 2)
    distance = math.sqrt(power)
    return distance



"""
Dictionaries module for Hft function to hub function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import hub.helpers.constants as cte


class HftFunctionToHubFunction:
  """
  Hft function to hub function class
  """
  def __init__(self):
    self._dictionary = {
      'residential': cte.RESIDENTIAL,
      'single family house': cte.SINGLE_FAMILY_HOUSE,
      'multifamily house': cte.MULTI_FAMILY_HOUSE,
      'hotel': cte.HOTEL,
      'hospital': cte.HOSPITAL,
      'outpatient': cte.OUT_PATIENT_HEALTH_CARE,
      'commercial': cte.SUPERMARKET,
      'strip mall': cte.STRIP_MALL,
      'warehouse': cte.WAREHOUSE,
      'primary school': cte.PRIMARY_SCHOOL,
      'secondary school': cte.EDUCATION,
      'office': cte.MEDIUM_OFFICE,
      'large office': cte.LARGE_OFFICE
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Non PV energy generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from typing import Union

from hub.city_model_structure.energy_systems.generation_system import GenerationSystem
from hub.city_model_structure.energy_systems.performance_curve import PerformanceCurves


class NonPvGenerationSystem(GenerationSystem):
  """
  NonPvGenerationSystem class
  """
  def __init__(self):
    super().__init__()
    self._nominal_heat_output = None
    self._maximum_heat_output = None
    self._minimum_heat_output = None
    self._heat_efficiency = None
    self._nominal_cooling_output = None
    self._maximum_cooling_output = None
    self._minimum_cooling_output = None
    self._cooling_efficiency = None
    self._electricity_efficiency = None
    self._nominal_electricity_output = None
    self._source_medium = None
    self._source_temperature = None
    self._source_mass_flow = None
    self._supply_medium = None
    self._maximum_heat_supply_temperature = None
    self._minimum_heat_supply_temperature = None
    self._maximum_cooling_supply_temperature = None
    self._minimum_cooling_supply_temperature = None
    self._heat_output_curve = None
    self._heat_fuel_consumption_curve = None
    self._heat_efficiency_curve = None
    self._cooling_output_curve = None
    self._cooling_fuel_consumption_curve = None
    self._cooling_efficiency_curve = None
    self._domestic_hot_water = None
    self._heat_supply_temperature = None
    self._cooling_supply_temperature = None
    self._reversible = None
    self._simultaneous_heat_cold = None
    self._energy_consumption = {}

  @property
  def nominal_heat_output(self):
    """
    Get nominal heat output of heat generation devices in W
    :return: float
    """
    return self._nominal_heat_output

  @nominal_heat_output.setter
  def nominal_heat_output(self, value):
    """
    Set nominal heat output of heat generation devices in W
    :param value: float
    """
    self._nominal_heat_output = value

  @property
  def maximum_heat_output(self):
    """
    Get maximum heat output of heat generation devices in W
    :return: float
    """
    return self._maximum_heat_output

  @maximum_heat_output.setter
  def maximum_heat_output(self, value):
    """
    Set maximum heat output of heat generation devices in W
    :param value: float
    """
    self._maximum_heat_output = value

  @property
  def minimum_heat_output(self):
    """
    Get minimum heat output of heat generation devices in W
    :return: float
    """
    return self._minimum_heat_output

  @minimum_heat_output.setter
  def minimum_heat_output(self, value):
    """
    Set minimum heat output of heat generation devices in W
    :param value: float
    """
    self._minimum_heat_output = value

  @property
  def source_medium(self):
    """
    Get source_type from [air, water, ground, district_heating, grid, on_site_electricity]
    :return: string
    """
    return self._source_medium

  @source_medium.setter
  def source_medium(self, value):
    """
    Set source medium from [Air, Water, Geothermal, District Heating, Grid, Onsite Electricity]
    :param value: [string]
    """
    self._source_medium = value

  @property
  def supply_medium(self):
    """
    Get the supply medium from ['air', 'water']
    :return: string
    """
    return self._supply_medium

  @supply_medium.setter
  def supply_medium(self, value):
    """
    Set the supply medium from ['air', 'water']
    :param value: string
    """
    self._supply_medium = value

  @property
  def heat_efficiency(self):
    """
    Get heat_efficiency
    :return: float
    """
    return self._heat_efficiency

  @heat_efficiency.setter
  def heat_efficiency(self, value):
    """
    Set heat_efficiency
    :param value: float
    """
    self._heat_efficiency = value

  @property
  def nominal_cooling_output(self):
    """
    Get nominal cooling output of heat generation devices in W
    :return: float
    """
    return self._nominal_cooling_output

  @nominal_cooling_output.setter
  def nominal_cooling_output(self, value):
    """
    Set nominal cooling output of heat generation devices in W
    :param value: float
    """
    self._nominal_cooling_output = value

  @property
  def maximum_cooling_output(self):
    """
    Get maximum heat output of heat generation devices in W
    :return: float
    """
    return self._maximum_cooling_output

  @maximum_cooling_output.setter
  def maximum_cooling_output(self, value):
    """
    Set maximum heat output of heat generation devices in W
    :param value: float
    """
    self._maximum_cooling_output = value

  @property
  def minimum_cooling_output(self):
    """
    Get minimum heat output of heat generation devices in W
    :return: float
    """
    return self._minimum_cooling_output

  @minimum_cooling_output.setter
  def minimum_cooling_output(self, value):
    """
    Set minimum heat output of heat generation devices in W
    :param value: float
    """
    self._minimum_cooling_output = value

  @property
  def cooling_efficiency(self):
    """
    Get cooling_efficiency
    :return: float
    """
    return self._cooling_efficiency

  @cooling_efficiency.setter
  def cooling_efficiency(self, value):
    """
    Set cooling_efficiency
    :param value: float
    """
    self._cooling_efficiency = value

  @property
  def electricity_efficiency(self):
    """
    Get electricity_efficiency
    :return: float
    """
    return self._electricity_efficiency

  @electricity_efficiency.setter
  def electricity_efficiency(self, value):
    """
    Set electricity_efficiency
    :param value: float
    """
    self._electricity_efficiency = value

  @property
  def source_temperature(self):
    """
    Get source_temperature in degree Celsius
    :return: float
    """
    return self._source_temperature

  @source_temperature.setter
  def source_temperature(self, value):
    """
    Set source_temperature in degree Celsius
    :param value: float
    """
    self._source_temperature = value

  @property
  def source_mass_flow(self):
    """
    Get source_mass_flow in kg/s
    :return: float
    """
    return self._source_mass_flow

  @source_mass_flow.setter
  def source_mass_flow(self, value):
    """
    Set source_mass_flow in kg/s
    :param value: float
    """
    self._source_mass_flow = value

  @property
  def nominal_electricity_output(self):
    """
    Get nominal_power_output of electricity generation devices or inverters in W
    :return: float
    """
    return self._nominal_electricity_output

  @nominal_electricity_output.setter
  def nominal_electricity_output(self, value):
    """
    Get nominal_power_output of electricity generation devices or inverters in W
    :param value: float
    """
    self._nominal_electricity_output = value

  @property
  def maximum_heat_supply_temperature(self):
    """
    Get the maximum heat supply temperature in degree Celsius
    :return: float
    """
    return self._minimum_heat_supply_temperature

  @maximum_heat_supply_temperature.setter
  def maximum_heat_supply_temperature(self, value):
    """
    Set maximum heating supply temperature in degree Celsius
    :param value: float
    """
    self._maximum_heat_supply_temperature = value

  @property
  def minimum_heat_supply_temperature(self):
    """
    Get the minimum heat supply temperature in degree Celsius
    :return: float
    """
    return self._minimum_heat_supply_temperature

  @minimum_heat_supply_temperature.setter
  def minimum_heat_supply_temperature(self, value):
    """
    Set minimum heating supply temperature in degree Celsius
    :param value: float
    """
    self._minimum_heat_supply_temperature = value

  @property
  def maximum_cooling_supply_temperature(self):
    """
    Get the maximum cooling supply temperature in degree Celsius
    :return: float
    """
    return self._maximum_cooling_supply_temperature

  @maximum_cooling_supply_temperature.setter
  def maximum_cooling_supply_temperature(self, value):
    """
    Set maximum cooling supply temperature in degree Celsius
    :param value: float
    """
    self._maximum_cooling_supply_temperature = value

  @property
  def minimum_cooling_supply_temperature(self):
    """
    Get the minimum cooling supply temperature in degree Celsius
    :return: float
    """
    return self._minimum_cooling_supply_temperature

  @minimum_cooling_supply_temperature.setter
  def minimum_cooling_supply_temperature(self, value):
    """
    Set minimum cooling supply temperature in degree Celsius
    :param value: float
    """
    self._minimum_cooling_supply_temperature = value

  @property
  def heat_output_curve(self) -> Union[None, PerformanceCurves]:
    """
    Get the heat output curve of the heat generation device
    :return: PerformanceCurve
    """
    return self._heat_output_curve

  @heat_output_curve.setter
  def heat_output_curve(self, value):
    """
    Set the heat output curve of the heat generation device
    :return: PerformanceCurve
    """
    self._heat_output_curve = value

  @property
  def heat_fuel_consumption_curve(self) -> Union[None, PerformanceCurves]:
    """
    Get the heating fuel consumption curve of the heat generation device
    :return: PerformanceCurve
    """
    return self._heat_fuel_consumption_curve

  @heat_fuel_consumption_curve.setter
  def heat_fuel_consumption_curve(self, value):
    """
    Set the heating fuel consumption curve of the heat generation device
    :return: PerformanceCurve
    """
    self._heat_fuel_consumption_curve = value

  @property
  def heat_efficiency_curve(self) -> Union[None, PerformanceCurves]:
    """
    Get the heating efficiency curve of the heat generation device
    :return: PerformanceCurve
    """
    return self._heat_efficiency_curve

  @heat_efficiency_curve.setter
  def heat_efficiency_curve(self, value):
    """
    Set the heating efficiency curve of the heat generation device
    :return: PerformanceCurve
    """
    self._heat_efficiency_curve = value

  @property
  def cooling_output_curve(self) -> Union[None, PerformanceCurves]:
    """
    Get the heat output curve of the heat generation device
    :return: PerformanceCurve
    """
    return self._cooling_output_curve

  @cooling_output_curve.setter
  def cooling_output_curve(self, value):
    """
    Set the cooling output curve of the heat generation device
    :return: PerformanceCurve
    """
    self._cooling_output_curve = value

  @property
  def cooling_fuel_consumption_curve(self) -> Union[None, PerformanceCurves]:
    """
    Get the heating fuel consumption curve of the heat generation device
    :return: PerformanceCurve
    """
    return self._cooling_fuel_consumption_curve

  @cooling_fuel_consumption_curve.setter
  def cooling_fuel_consumption_curve(self, value):
    """
    Set the heating fuel consumption curve of the heat generation device
    :return: PerformanceCurve
    """
    self._cooling_fuel_consumption_curve = value

  @property
  def cooling_efficiency_curve(self) -> Union[None, PerformanceCurves]:
    """
    Get the heating efficiency curve of the heat generation device
    :return: PerformanceCurve
    """
    return self._cooling_efficiency_curve

  @cooling_efficiency_curve.setter
  def cooling_efficiency_curve(self, value):
    """
    Set the heating efficiency curve of the heat generation device
    :return: PerformanceCurve
    """
    self._cooling_efficiency_curve = value

  @property
  def domestic_hot_water(self):
    """
    Get the capability of generating domestic hot water

    :return: bool
    """
    return self._domestic_hot_water

  @domestic_hot_water.setter
  def domestic_hot_water(self, value):
    """
    Set the capability of generating domestic hot water

    :return: bool
    """
    self._domestic_hot_water = value

  @property
  def heat_supply_temperature(self):
    """
    Get the hourly heat supply temperature
    :return: list
    """
    return self._heat_supply_temperature

  @heat_supply_temperature.setter
  def heat_supply_temperature(self, value):
    """
    set the hourly heat supply temperature
    :param value:
    :return: list
    """
    self._heat_supply_temperature = value

  @property
  def cooling_supply_temperature(self):
    """
    Get the hourly cooling supply temperature
    :return: list
    """
    return self._heat_supply_temperature

  @cooling_supply_temperature.setter
  def cooling_supply_temperature(self, value):
    """
    set the hourly cooling supply temperature
    :param value:
    :return: list
    """
    self._cooling_supply_temperature = value

  @property
  def reversibility(self):
    """
    Get the capability of generating both heating and cooling

    :return: bool
    """
    return self._reversible

  @reversibility.setter
  def reversibility(self, value):
    """
    Set the capability of generating domestic hot water

    :return: bool
    """
    self._reversible = value

  @property
  def simultaneous_heat_cold(self):
    """
    Get the capability of generating both heating and cooling at the same time

    :return: bool
    """
    return self._simultaneous_heat_cold

  @simultaneous_heat_cold.setter
  def simultaneous_heat_cold(self, value):
    """
    Set the capability of generating domestic hot water at the same time

    :return: bool
    """
    self._simultaneous_heat_cold = value

  @property
  def energy_consumption(self) -> dict:
    """
    Get energy consumption in W
    :return: dict{[float]}
    """
    return self._energy_consumption

  @energy_consumption.setter
  def energy_consumption(self, value):
    """
    Set energy consumption in W
    :param value: dict{[float]}
    """
    self._energy_consumption = value




"""
ConstructionFactory (before PhysicsFactory) retrieve the specific construction module for the given region
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.helpers.utils import validate_import_export_type
from hub.imports.construction.nrcan_physics_parameters import NrcanPhysicsParameters
from hub.imports.construction.nrel_physics_parameters import NrelPhysicsParameters
from hub.imports.construction.eilat_physics_parameters import EilatPhysicsParameters
from hub.imports.construction.palma_physics_parameters import PalmaPhysicsParameters


class ConstructionFactory:
  """
  ConstructionFactory class
  """
  def __init__(self, handler, city):
    self._handler = '_' + handler.lower()
    validate_import_export_type(ConstructionFactory, handler)
    self._city = city

  def _nrel(self):
    """
    Enrich the city by using NREL information
    """
    NrelPhysicsParameters(self._city).enrich_buildings()
    self._city.level_of_detail.construction = 2
    for building in self._city.buildings:
      building.level_of_detail.construction = 2

  def _nrcan(self):
    """
    Enrich the city by using NRCAN information
    """
    NrcanPhysicsParameters(self._city).enrich_buildings()
    self._city.level_of_detail.construction = 2
    for building in self._city.buildings:
      building.level_of_detail.construction = 2

  def _eilat(self):
    """
    Enrich the city by using Eilat information
    """
    EilatPhysicsParameters(self._city).enrich_buildings()
    self._city.level_of_detail.construction = 2
    for building in self._city.buildings:
      building.level_of_detail.construction = 2

  def _palma(self):
    """
    Enrich the city by using Palma information
    """
    PalmaPhysicsParameters(self._city).enrich_buildings()
    self._city.level_of_detail.construction = 2
    for building in self._city.buildings:
      building.level_of_detail.construction = 2

  def enrich(self):
    """
    Enrich the city given to the class using the class given handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()
    for building in self._city.buildings:
      _ = building.thermal_zones_from_internal_zones  # ensure internal zones initialization



"""
Energy System catalog emission system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class EmissionSystem:
  """
  Emission system class
  """
  def __init__(self, system_id, model_name=None, system_type=None, parasitic_energy_consumption=0):

    self._system_id = system_id
    self._model_name = model_name
    self._type = system_type
    self._parasitic_energy_consumption = parasitic_energy_consumption

  @property
  def id(self):
    """
    Get system id
    :return: float
    """
    return self._system_id

  @property
  def model_name(self):
    """
    Get model name
    :return: string
    """
    return self._model_name

  @property
  def type(self):
    """
    Get type
    :return: string
    """
    return self._type

  @property
  def parasitic_energy_consumption(self):
    """
    Get parasitic_energy_consumption in ratio (J/J)
    :return: float
    """
    return self._parasitic_energy_consumption

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Layer': {'id': self.id,
                         'model name': self.model_name,
                         'type': self.type,
                         'parasitic energy consumption per energy produced [J/J]': self.parasitic_energy_consumption
                         }
               }
    return content



"""
Building module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
from typing import List, Union, TypeVar

import numpy as np
import hub.helpers.constants as cte
from hub.city_model_structure.attributes.polyhedron import Polyhedron
from hub.city_model_structure.building_demand.household import Household
from hub.city_model_structure.building_demand.internal_zone import InternalZone
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.helpers.peak_loads import PeakLoads

City = TypeVar('City')


class Building(CityObject):
  """
  Building(CityObject) class
  """
  def __init__(self, name, surfaces, year_of_construction, function, usages=None, terrains=None, city=None):
    super().__init__(name, surfaces)
    self._city = city
    self._households = None
    self._basement_heated = None
    self._attic_heated = None
    self._terrains = terrains
    self._year_of_construction = year_of_construction
    self._function = function
    self._usages = usages
    self._average_storey_height = None
    self._storeys_above_ground = None
    self._floor_area = None
    self._roof_type = None
    self._internal_zones = None
    self._thermal_zones_from_internal_zones = None
    self._shell = None
    self._aliases = []
    self._type = 'building'
    self._cold_water_temperature = {}
    self._heating_demand = {}
    self._cooling_demand = {}
    self._lighting_electrical_demand = {}
    self._appliances_electrical_demand = {}
    self._domestic_hot_water_heat_demand = {}
    self._heating_consumption = {}
    self._cooling_consumption = {}
    self._domestic_hot_water_consumption = {}
    self._distribution_systems_electrical_consumption = {}
    self._onsite_electrical_production = {}
    self._eave_height = None
    self._energy_systems = None
    self._systems_archetype_name = None
    self._grounds = []
    self._roofs = []
    self._walls = []
    self._internal_walls = []
    self._ground_walls = []
    self._attic_floors = []
    self._interior_slabs = []
    for surface_id, surface in enumerate(self.surfaces):
      self._min_x = min(self._min_x, surface.lower_corner[0])
      self._min_y = min(self._min_y, surface.lower_corner[1])
      self._min_z = min(self._min_z, surface.lower_corner[2])
      self._max_x = max(self._max_x, surface.upper_corner[0])
      self._max_y = max(self._max_y, surface.upper_corner[1])
      self._max_z = max(self._max_z, surface.upper_corner[2])
      surface.id = surface_id
      if surface.type == cte.GROUND:
        self._grounds.append(surface)
      elif surface.type == cte.WALL:
        self._walls.append(surface)
      elif surface.type == cte.ROOF:
        self._roofs.append(surface)
      elif surface.type == cte.INTERIOR_WALL:
        self._internal_walls.append(surface)
      elif surface.type == cte.GROUND_WALL:
        self._ground_walls.append(surface)
      elif surface.type == cte.ATTIC_FLOOR:
        self._attic_floors.append(surface)
      elif surface.type == cte.INTERIOR_SLAB:
        self._interior_slabs.append(surface)
      else:
        logging.error('Building %s [%s] has an unexpected surface type %s.', self.name, self.aliases, surface.type)
    self._domestic_hot_water_peak_load = None
    self._fuel_consumption_breakdown = {}
    self._systems_archetype_cluster_id = None
    self._pv_generation = {}

  @property
  def shell(self) -> Polyhedron:
    """
    Get building's external polyhedron
    :return: [Polyhedron]
    """
    polygons = []
    for surface in self.surfaces:
      if surface.type is not cte.INTERIOR_WALL:
        polygons.append(surface.solid_polygon)
        if surface.holes_polygons is not None:
          for hole in surface.holes_polygons:
            polygons.append(hole)
    if self._shell is None:
      self._shell = Polyhedron(polygons)
    return self._shell

  @property
  def internal_zones(self) -> List[InternalZone]:
    """
    Get building internal zones
    For Lod up to 3, there is only one internal zone which corresponds to the building shell.
    In LoD 4 there can be more than one. In this case the definition of surfaces and floor area must be redefined.
    :return: [InternalZone]
    """
    if self._internal_zones is None:
      self._internal_zones = [InternalZone(self.surfaces, self.floor_area, self.volume)]
    return self._internal_zones

  @property
  def thermal_zones_from_internal_zones(self) -> Union[None, List[ThermalZone]]:
    """
    Get building thermal zones
    :return: [ThermalZone]
    """
    if self._thermal_zones_from_internal_zones is None:
      self._thermal_zones_from_internal_zones = []
      for internal_zone in self.internal_zones:
        if internal_zone.thermal_zones_from_internal_zones is None:
          self._thermal_zones_from_internal_zones = None
          return self._thermal_zones_from_internal_zones
        self._thermal_zones_from_internal_zones.append(internal_zone.thermal_zones_from_internal_zones[0])
    return self._thermal_zones_from_internal_zones

  @property
  def grounds(self) -> List[Surface]:
    """
    Get building ground surfaces
    :return: [Surface]
    """
    return self._grounds

  @property
  def roofs(self) -> List[Surface]:
    """
    Get building roof surfaces
    :return: [Surface]
    """
    return self._roofs

  @property
  def walls(self) -> List[Surface]:
    """
    Get building wall surfaces
    :return: [Surface]
    """
    return self._walls

  @property
  def internal_walls(self) -> List[Surface]:
    """
    Get building internal wall surfaces
    :return: [Surface]
    """
    return self._internal_walls

  @property
  def terrains(self) -> Union[None, List[Surface]]:
    """
    Get city object terrain surfaces
    :return: [Surface]
    """
    return self._terrains

  @property
  def attic_heated(self) -> Union[None, int]:
    """
    Get if the city object attic is heated
    0: no attic in the building
    1: attic exists but is not heated
    2: attic exists and is heated
    :return: None or int
    """
    return self._attic_heated

  @attic_heated.setter
  def attic_heated(self, value):
    """
    Set if the city object attic is heated
    0: no attic in the building
    1: attic exists but is not heated
    2: attic exists and is heated
    :param value: int
    """
    if value is not None:
      self._attic_heated = int(value)

  @property
  def basement_heated(self) -> Union[None, int]:
    """
    Get if the city object basement is heated
    0: no basement in the building
    1: basement exists but is not heated
    2: basement exists and is heated
    :return: None or int
    """
    return self._basement_heated

  @basement_heated.setter
  def basement_heated(self, value):
    """
    Set if the city object basement is heated
    0: no basement in the building
    1: basement exists but is not heated
    2: basement exists and is heated
    :param value: int
    """
    if value is not None:
      self._basement_heated = int(value)

  @property
  def year_of_construction(self):
    """
    Get building year of construction
    :return: int
    """
    return self._year_of_construction

  @year_of_construction.setter
  def year_of_construction(self, value):
    """
    Set building year of construction
    :param value: int
    """
    if value is not None:
      self._year_of_construction = int(value)

  @property
  def function(self) -> Union[None, str]:
    """
    Get building function
    :return: None or str
    """
    return self._function

  @function.setter
  def function(self, value):
    """
    Set building function
    :param value: str
    """
    if value is not None:
      self._function = value

  @property
  def usages(self) -> Union[None, list]:
    """
    Get building usages, if none, assume usage is function
    :return: None or list of functions
    """
    if self._usages is None and self._function is not None:
      self._usages = [{'usage': self._function, 'ratio': 1 }]
    return self._usages

  @property
  def average_storey_height(self) -> Union[None, float]:
    """
    Get building average storey height in meters
    :return: None or float
    """
    if len(self.internal_zones) > 1:
      self._average_storey_height = 0
      for internal_zone in self.internal_zones:
        self._average_storey_height += internal_zone.mean_height / len(self.internal_zones)
    else:
      if self.internal_zones[0].thermal_archetype is None:
        self._average_storey_height = None
      else:
        self._average_storey_height = self.internal_zones[0].thermal_archetype.average_storey_height
    return self._average_storey_height

  @average_storey_height.setter
  def average_storey_height(self, value):
    """
    Set building average storey height in meters
    :param value: float
    """
    if value is not None:
      self._average_storey_height = float(value)

  @property
  def storeys_above_ground(self) -> Union[None, int]:
    """
    Get building storeys number above ground
    :return: None or int
    """
    if self._storeys_above_ground is None:
      if self.eave_height is not None and self.average_storey_height is not None:
        storeys_above_ground = int(self.eave_height / self.average_storey_height)
        if storeys_above_ground == 0:
          storeys_above_ground += 1
        self._storeys_above_ground = storeys_above_ground
    return self._storeys_above_ground

  @storeys_above_ground.setter
  def storeys_above_ground(self, value):
    """
    Set building storeys number above ground
    :param value: int
    """
    if value is not None:
      self._storeys_above_ground = int(value)

  @property
  def cold_water_temperature(self) -> {float}:
    """
    Get cold water temperature in degrees Celsius
    :return: dict{[float]}
    """
    return self._cold_water_temperature

  @cold_water_temperature.setter
  def cold_water_temperature(self, value):
    """
    Set cold water temperature in degrees Celsius
    :param value: dict{[float]}
    """
    self._cold_water_temperature = value

  @property
  def heating_demand(self) -> dict:
    """
    Get heating demand in J
    :return: dict{[float]}
    """
    return self._heating_demand

  @heating_demand.setter
  def heating_demand(self, value):
    """
    Set heating demand in J
    :param value: dict{[float]}
    """
    self._heating_demand = value

  @property
  def cooling_demand(self) -> dict:
    """
    Get cooling demand in J
    :return: dict{[float]}
    """
    return self._cooling_demand

  @cooling_demand.setter
  def cooling_demand(self, value):
    """
    Set cooling demand in J
    :param value: dict{[float]}
    """
    self._cooling_demand = value

  @property
  def lighting_electrical_demand(self) -> dict:
    """
    Get lighting electrical demand in J
    :return: dict{[float]}
    """
    return self._lighting_electrical_demand

  @lighting_electrical_demand.setter
  def lighting_electrical_demand(self, value):
    """
    Set lighting electrical demand in J
    :param value: dict{[float]}
    """
    self._lighting_electrical_demand = value

  @property
  def appliances_electrical_demand(self) -> dict:
    """
    Get appliances electrical demand in J
    :return: dict{[float]}
    """
    return self._appliances_electrical_demand

  @appliances_electrical_demand.setter
  def appliances_electrical_demand(self, value):
    """
    Set appliances electrical demand in J
    :param value: dict{[float]}
    """
    self._appliances_electrical_demand = value

  @property
  def domestic_hot_water_heat_demand(self) -> dict:
    """
    Get domestic hot water heat demand in J
    :return: dict{[float]}
    """
    return self._domestic_hot_water_heat_demand

  @domestic_hot_water_heat_demand.setter
  def domestic_hot_water_heat_demand(self, value):
    """
    Set domestic hot water heat demand in J
    :param value: dict{[float]}
    """
    self._domestic_hot_water_heat_demand = value

  @property
  def lighting_peak_load(self) -> Union[None, dict]:
    """
    Get lighting peak load in W
    :return: dict{[float]}
    """
    results = {}
    peak_lighting = 0
    peak = 0
    for thermal_zone in self.thermal_zones_from_internal_zones:
      lighting = thermal_zone.lighting
      for schedule in lighting.schedules:
        peak = max(schedule.values) * lighting.density * thermal_zone.total_floor_area
        if peak > peak_lighting:
          peak_lighting = peak
    results[cte.MONTH] = [peak for _ in range(0, 12)]
    results[cte.YEAR] = [peak]
    return results

  @property
  def appliances_peak_load(self) -> Union[None, dict]:
    """
    Get appliances peak load in W
    :return: dict{[float]}
    """
    results = {}
    peak_appliances = 0
    peak = 0
    for thermal_zone in self.thermal_zones_from_internal_zones:
      appliances = thermal_zone.appliances
      for schedule in appliances.schedules:
        peak = max(schedule.values) * appliances.density * thermal_zone.total_floor_area
        if peak > peak_appliances:
          peak_appliances = peak
    results[cte.MONTH] = [peak for _ in range(0, 12)]
    results[cte.YEAR] = [peak]
    return results

  @property
  def heating_peak_load(self) -> Union[None, dict]:
    """
    Get heating peak load in W
    :return: dict{[float]}
    """
    results = {}
    if cte.HOUR in self.heating_demand:
      monthly_values = PeakLoads().peak_loads_from_hourly(self.heating_demand[cte.HOUR])
    else:
      monthly_values = PeakLoads(self).heating_peak_loads_from_methodology
    if monthly_values is None:
      return None
    results[cte.MONTH] = [x / cte.WATTS_HOUR_TO_JULES for x in monthly_values]
    results[cte.YEAR] = [max(monthly_values) / cte.WATTS_HOUR_TO_JULES]
    return results

  @property
  def cooling_peak_load(self) -> Union[None, dict]:
    """
    Get cooling peak load in W
    :return: dict{[float]}
    """
    results = {}
    if cte.HOUR in self.cooling_demand:
      monthly_values = PeakLoads().peak_loads_from_hourly(self.cooling_demand[cte.HOUR])
    else:
      monthly_values = PeakLoads(self).cooling_peak_loads_from_methodology
    if monthly_values is None:
      return None
    results[cte.MONTH] = [x / cte.WATTS_HOUR_TO_JULES for x in monthly_values]
    results[cte.YEAR] = [max(monthly_values) / cte.WATTS_HOUR_TO_JULES]
    return results

  @property
  def domestic_hot_water_peak_load(self) -> Union[None, dict]:
    """
    Get cooling peak load in W
    :return: dict{[float]}
    """
    results = {}
    monthly_values = None
    if cte.HOUR in self.domestic_hot_water_heat_demand:
      monthly_values = PeakLoads().peak_loads_from_hourly(self.domestic_hot_water_heat_demand[cte.HOUR])
    if monthly_values is None:
      return None
    results[cte.MONTH] = [x / cte.WATTS_HOUR_TO_JULES for x in monthly_values]
    results[cte.YEAR] = [max(monthly_values) / cte.WATTS_HOUR_TO_JULES]
    return results

  @property
  def eave_height(self):
    """
    Get building eave height in meters
    :return: float
    """
    if self._eave_height is None:
      self._eave_height = 0
      for wall in self.walls:
        self._eave_height = max(self._eave_height, wall.upper_corner[2]) - self.simplified_polyhedron.min_z
    return self._eave_height

  @property
  def roof_type(self):
    """
    Get roof type for the building flat or pitch
    :return: str
    """
    if self._roof_type is None:
      self._roof_type = 'flat'
      for roof in self.roofs:
        grads = np.rad2deg(roof.inclination)
        if 355 > grads > 5:
          self._roof_type = 'pitch'
          break
    return self._roof_type

  @roof_type.setter
  def roof_type(self, value):
    """
    Set roof type for the building flat or pitch
    :return: str
    """
    self._roof_type = value

  @property
  def floor_area(self):
    """
    Get building floor area in square meters
    :return: float
    """
    if self._floor_area is None:
      self._floor_area = 0
      for surface in self.surfaces:
        if surface.type == 'Ground':
          self._floor_area += surface.perimeter_polygon.area
    return self._floor_area

  @property
  def households(self) -> List[Household]:
    """
    Get the list of households inside the building
    :return: List[Household]
    """
    return self._households

  @property
  def is_conditioned(self):
    """
    Get building heated flag
    :return: Boolean
    """
    if self.internal_zones is None:
      return False
    for internal_zone in self.internal_zones:
      if internal_zone.usages is not None:
        for usage in internal_zone.usages:
          if usage.thermal_control is not None:
            return True
    return False

  @property
  def aliases(self):
    """
    Get the alias name for the building
    :return: str
    """
    return self._aliases

  def add_alias(self, value):
    """
    Add a new alias for the building
    """
    self._aliases.append(value)
    if self.city is not None:
      self.city.add_building_alias(self, value)

  @property
  def city(self) -> City:
    """
    Get the city containing the building
    :return: City
    """
    return self._city

  @city.setter
  def city(self, value):
    """
    Set the city containing the building
    """
    self._city = value

  @property
  def energy_systems(self) -> Union[None, List[EnergySystem]]:
    """
    Get list of energy systems installed to cover the building demands
    :return: [EnergySystem]
    """
    return self._energy_systems

  @energy_systems.setter
  def energy_systems(self, value):
    """
    Set list of energy systems installed to cover the building demands
    :param value: [EnergySystem]
    """
    self._energy_systems = value

  @property
  def energy_systems_archetype_name(self):
    """
    Get energy systems archetype name
    :return: str
    """
    return self._systems_archetype_name

  @energy_systems_archetype_name.setter
  def energy_systems_archetype_name(self, value):
    """
    Set energy systems archetype name
    :param value: str
    """
    self._systems_archetype_name = value

  @property
  def heating_consumption(self):
    """
    Get energy consumption for heating according to the heating system installed in J
    return: dict
    """
    if len(self._heating_consumption) == 0:
      for heating_demand_key in self.heating_demand:
        demand = self.heating_demand[heating_demand_key]
        consumption_type = cte.HEATING
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._heating_consumption[heating_demand_key] = final_energy_consumed
    return self._heating_consumption

  @property
  def cooling_consumption(self):
    """
    Get energy consumption for cooling according to the cooling system installed in J
    return: dict
    """
    if len(self._cooling_consumption) == 0:
      for cooling_demand_key in self.cooling_demand:
        demand = self.cooling_demand[cooling_demand_key]
        consumption_type = cte.COOLING
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._cooling_consumption[cooling_demand_key] = final_energy_consumed
    return self._cooling_consumption

  @property
  def domestic_hot_water_consumption(self):
    """
    Get energy consumption for domestic according to the domestic hot water system installed in J
    return: dict
    """
    if len(self._domestic_hot_water_consumption) == 0:
      for domestic_hot_water_demand_key in self.domestic_hot_water_heat_demand:
        demand = self.domestic_hot_water_heat_demand[domestic_hot_water_demand_key]
        consumption_type = cte.DOMESTIC_HOT_WATER
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._domestic_hot_water_consumption[domestic_hot_water_demand_key] = final_energy_consumed
    return self._domestic_hot_water_consumption

  def _calculate_working_hours(self):
    _working_hours = {}
    for internal_zone in self.internal_zones:
      for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
        _working_hours_per_thermal_zone = {}
        for schedule in thermal_zone.thermal_control.hvac_availability_schedules:
          _working_hours_per_schedule = [0] * len(schedule.values)
          for i, value in enumerate(schedule.values):
            if value > 0:
              _working_hours_per_schedule[i] = 1
          for day_type in schedule.day_types:
            _working_hours_per_thermal_zone[day_type] = _working_hours_per_schedule
        if len(_working_hours) == 0:
          _working_hours = _working_hours_per_thermal_zone
        else:
          for key, item in _working_hours.items():
            saved_values = _working_hours_per_thermal_zone[key]
            for i, value in enumerate(item):
              _working_hours[key][i] = max(_working_hours[key][i], saved_values[i])

    working_hours = {}
    values_months = []
    for month in cte.WEEK_DAYS_A_MONTH.keys():
      _total_hours_month = 0
      for key in _working_hours:
        hours = sum(_working_hours[key])
        _total_hours_month += hours * cte.WEEK_DAYS_A_MONTH[month][key]
      values_months.append(_total_hours_month)
    working_hours[cte.MONTH] = values_months
    working_hours[cte.YEAR] = sum(working_hours[cte.MONTH])
    return working_hours

  @property
  def distribution_systems_electrical_consumption(self):
    """
    Get total electricity consumption for distribution and emission systems in J
    return: dict
    """
    _distribution_systems_electrical_consumption = {}
    if len(self._distribution_systems_electrical_consumption) != 0:
      return self._distribution_systems_electrical_consumption
    _peak_load = self.heating_peak_load[cte.YEAR][0]
    _peak_load_type = cte.HEATING
    if _peak_load < self.cooling_peak_load[cte.YEAR][0]:
      _peak_load = self.cooling_peak_load[cte.YEAR][0]
      _peak_load_type = cte.COOLING

    _working_hours = self._calculate_working_hours()
    _consumption_fix_flow = 0
    if self.energy_systems is None:
      return self._distribution_systems_electrical_consumption
    for energy_system in self.energy_systems:
      distribution_systems = energy_system.distribution_systems
      if distribution_systems is not None:
        for distribution_system in distribution_systems:
          emission_systems = distribution_system.emission_systems
          parasitic_energy_consumption = 0
          if emission_systems is not None:
            for emission_system in emission_systems:
              parasitic_energy_consumption += emission_system.parasitic_energy_consumption
          consumption_variable_flow = distribution_system.distribution_consumption_variable_flow
          for demand_type in energy_system.demand_types:
            if demand_type.lower() == cte.HEATING.lower():
              if _peak_load_type == cte.HEATING.lower():
                _consumption_fix_flow = distribution_system.distribution_consumption_fix_flow
              for heating_demand_key in self.heating_demand:
                _consumption = [0]*len(self.heating_demand[heating_demand_key])
                _demand = self.heating_demand[heating_demand_key]
                for i, _ in enumerate(_consumption):
                  _consumption[i] += (parasitic_energy_consumption + consumption_variable_flow) * _demand[i]
                self._distribution_systems_electrical_consumption[heating_demand_key] = _consumption
            if demand_type.lower() == cte.COOLING.lower():
              if _peak_load_type == cte.COOLING.lower():
                _consumption_fix_flow = distribution_system.distribution_consumption_fix_flow
              for demand_key in self.cooling_demand:
                _consumption = self._distribution_systems_electrical_consumption[demand_key]
                _demand = self.cooling_demand[demand_key]
                for i, _ in enumerate(_consumption):
                  _consumption[i] += (parasitic_energy_consumption + consumption_variable_flow) * _demand[i]
                self._distribution_systems_electrical_consumption[demand_key] = _consumption

        for key, item in self._distribution_systems_electrical_consumption.items():
          for i in range(0, len(item)):
            _working_hours_value = _working_hours[key]
            if len(item) == 12:
              _working_hours_value = _working_hours[key][i]
            self._distribution_systems_electrical_consumption[key][i] += (
              _peak_load * _consumption_fix_flow * _working_hours_value * cte.WATTS_HOUR_TO_JULES
            )

    return self._distribution_systems_electrical_consumption

  def _calculate_consumption(self, consumption_type, demand):
    # todo: modify when COP depends on the hour
    coefficient_of_performance = 0
    if self.energy_systems is None:
      return None
    for energy_system in self.energy_systems:
      generation_systems = energy_system.generation_systems
      for demand_type in energy_system.demand_types:
        if demand_type.lower() == consumption_type.lower():
          if consumption_type in (cte.HEATING, cte.DOMESTIC_HOT_WATER):
            for generation_system in generation_systems:
              if generation_system.heat_efficiency is not None:
                coefficient_of_performance = float(generation_system.heat_efficiency)
          elif consumption_type == cte.COOLING:
            for generation_system in generation_systems:
              if generation_system.cooling_efficiency is not None:
                coefficient_of_performance = float(generation_system.cooling_efficiency)
          elif consumption_type == cte.ELECTRICITY:
            for generation_system in generation_systems:
              if generation_system.electricity_efficiency is not None:
                coefficient_of_performance = float(generation_system.electricity_efficiency)
    if coefficient_of_performance == 0:
      values = [0]*len(demand)
      final_energy_consumed = values
    else:
      final_energy_consumed = []
      for demand_value in demand:
        final_energy_consumed.append(demand_value / coefficient_of_performance)
    return final_energy_consumed

  @property
  def onsite_electrical_production(self):
    """
    Get total electricity produced onsite in J
    return: dict
    """
    orientation_losses_factor = {cte.MONTH: {'north': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                             'east': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                             'south': [2.137931, 1.645503, 1.320946, 1.107817, 0.993213, 0.945175,
                                                       0.967949, 1.065534, 1.24183, 1.486486, 1.918033, 2.210526],
                                             'west': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]},
                                 cte.YEAR: {'north': [0],
                                            'east': [0],
                                            'south': [1.212544],
                                            'west': [0]}
                                 }

    # Add other systems whenever new ones appear
    if self.energy_systems is None:
      return self._onsite_electrical_production
    for energy_system in self.energy_systems:
      for generation_system in energy_system.generation_systems:
        if generation_system.system_type == cte.PHOTOVOLTAIC:
          if generation_system.electricity_efficiency is not None:
            _efficiency = float(generation_system.electricity_efficiency)
          else:
            _efficiency = 0
          self._onsite_electrical_production = {}
          for _key in self.roofs[0].global_irradiance.keys():
            _results = [0 for _ in range(0, len(self.roofs[0].global_irradiance[_key]))]
            for surface in self.roofs:
              if _key in orientation_losses_factor:
                _results = [x + y * _efficiency * surface.perimeter_area
                            * surface.solar_collectors_area_reduction_factor * z
                            for x, y, z in zip(_results, surface.global_irradiance[_key],
                                               orientation_losses_factor[_key]['south'])]
            self._onsite_electrical_production[_key] = _results
    return self._onsite_electrical_production

  @property
  def lower_corner(self):
    """
    Get building lower corner.
    """
    return [self._min_x, self._min_y, self._min_z]

  @property
  def upper_corner(self):
    """
    Get building upper corner.
    """
    return [self._max_x, self._max_y, self._max_z]

  @property
  def energy_consumption_breakdown(self) -> dict:
    """
    Get energy consumption of different sectors
    return: dict
    """
    fuel_breakdown = {cte.ELECTRICITY: {cte.LIGHTING: self.lighting_electrical_demand[cte.YEAR][0] if self.lighting_electrical_demand else 0,
                                        cte.APPLIANCES: self.appliances_electrical_demand[cte.YEAR][0] if self.appliances_electrical_demand else 0}}
    energy_systems = self.energy_systems
    if energy_systems is not None:
      for energy_system in energy_systems:
        demand_types = energy_system.demand_types
        generation_systems = energy_system.generation_systems
        for demand_type in demand_types:
          for generation_system in generation_systems:
            if generation_system.system_type != cte.PHOTOVOLTAIC:
              if generation_system.fuel_type not in fuel_breakdown:
                fuel_breakdown[generation_system.fuel_type] = {}
              if demand_type in generation_system.energy_consumption:
                fuel_breakdown[f'{generation_system.fuel_type}'][f'{demand_type}'] = (
                  generation_system.energy_consumption)[f'{demand_type}'][cte.YEAR][0]
              storage_systems = generation_system.energy_storage_systems
              if storage_systems:
                for storage_system in storage_systems:
                  if storage_system.type_energy_stored == 'thermal' and storage_system.heating_coil_energy_consumption:
                    fuel_breakdown[cte.ELECTRICITY][f'{demand_type}'] += (
                      storage_system.heating_coil_energy_consumption)[f'{demand_type}'][cte.YEAR][0]
      #TODO: When simulation models of all energy system archetypes are created, this part can be removed
      heating_fuels = []
      dhw_fuels = []
      for energy_system in self.energy_systems:
        if cte.HEATING in energy_system.demand_types:
          for generation_system in energy_system.generation_systems:
            heating_fuels.append(generation_system.fuel_type)
        if cte.DOMESTIC_HOT_WATER in energy_system.demand_types:
          for generation_system in energy_system.generation_systems:
            dhw_fuels.append(generation_system.fuel_type)
      for key in fuel_breakdown:
        if key == cte.ELECTRICITY and cte.COOLING not in fuel_breakdown[key]:
          for energy_system in energy_systems:
            if cte.COOLING in energy_system.demand_types and cte.COOLING not in fuel_breakdown[key]:
              if self.cooling_consumption:
                fuel_breakdown[energy_system.generation_systems[0].fuel_type][cte.COOLING] = self.cooling_consumption[cte.YEAR][0]
        for fuel in heating_fuels:
          if cte.HEATING not in fuel_breakdown[fuel]:
            for energy_system in energy_systems:
              if cte.HEATING in energy_system.demand_types:
                if self.heating_consumption:
                  fuel_breakdown[energy_system.generation_systems[0].fuel_type][cte.HEATING] = self.heating_consumption[cte.YEAR][0]
        for fuel in dhw_fuels:
          if cte.DOMESTIC_HOT_WATER not in fuel_breakdown[fuel]:
            for energy_system in energy_systems:
              if cte.DOMESTIC_HOT_WATER in energy_system.demand_types:
                if self.domestic_hot_water_consumption:
                  fuel_breakdown[energy_system.generation_systems[0].fuel_type][cte.DOMESTIC_HOT_WATER] = self.domestic_hot_water_consumption[cte.YEAR][0]
    self._fuel_consumption_breakdown = fuel_breakdown
    return self._fuel_consumption_breakdown

  @property
  def energy_systems_archetype_cluster_id(self):
    """
    Get energy systems archetype id
    :return: str
    """
    return self._systems_archetype_cluster_id

  @energy_systems_archetype_cluster_id.setter
  def energy_systems_archetype_cluster_id(self, value):
    """
    Set energy systems archetype id
    :param value: str
    """
    self._systems_archetype_cluster_id = value

  @property
  def pv_generation(self):
    """
    temporary attribute to get the onsite pv generation in W
    :return: dict
    """
    return self._pv_generation

  @pv_generation.setter
  def pv_generation(self, value):
    """
    temporary attribute to set the onsite pv generation in W
    :param value: float
    """
    self._pv_generation = value




"""
InternalZone module. It saves the original geometrical information from interiors together with some attributes of those
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import uuid
from typing import Union, List
from hub.city_model_structure.building_demand.usage import Usage
from hub.city_model_structure.building_demand.thermal_archetype import ThermalArchetype
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone
from hub.city_model_structure.building_demand.thermal_boundary import ThermalBoundary
from hub.city_model_structure.attributes.polyhedron import Polyhedron


class InternalZone:
  """
  InternalZone class
  """
  def __init__(self, surfaces, area, volume):
    self._surfaces = surfaces
    self._id = None
    self._geometry = None
    self._volume = volume
    self._area = area
    self._thermal_zones_from_internal_zones = None
    self._usages = None
    self._thermal_archetype = None

  @property
  def id(self):
    """
    Get internal zone id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def geometry(self) -> Polyhedron:
    """
    Get internal zone geometry
    :return: Polyhedron
    """
    if self._geometry is None:
      polygons = []
      for surface in self.surfaces:
        polygons.append(surface.perimeter_polygon)
      self._geometry = Polyhedron(polygons)
    return self._geometry

  @property
  def surfaces(self):
    """
    Get internal zone surfaces
    :return: [Surface]
    """
    return self._surfaces

  @property
  def volume(self):
    """
    Get internal zone volume in cubic meters
    :return: float
    """
    return self._volume

  @property
  def area(self):
    """
    Get internal zone area in square meters
    :return: float
    """
    return self._area

  @property
  def mean_height(self):
    """
    Get internal zone mean height in meters
    :return: float
    """
    return self.volume / self.area

  @property
  def usages(self) -> [Usage]:
    """
    Get usage archetypes
    :return: [Usage]
    """
    return self._usages

  @usages.setter
  def usages(self, value):
    """
    Set usage archetypes
    :param value: [Usage]
    """
    self._usages = value

  @property
  def thermal_archetype(self) -> ThermalArchetype:
    """
    Get thermal archetype parameters
    :return: ThermalArchetype
    """
    return self._thermal_archetype

  @thermal_archetype.setter
  def thermal_archetype(self, value):
    """
    Set thermal archetype parameters
    :param value: ThermalArchetype
    """
    self._thermal_archetype = value

  @property
  def thermal_zones_from_internal_zones(self) -> Union[None, List[ThermalZone]]:
    """
    Get building thermal zones as one per internal zone
    :return: [ThermalZone]
    """
    _thermal_boundaries = []
    for surface in self.surfaces:
      if surface.holes_polygons is None:
        windows_areas = None
      else:
        windows_areas = []
        for hole in surface.holes_polygons:
          windows_areas.append(hole.area)
      _thermal_boundary = ThermalBoundary(surface, surface.solid_polygon.area, windows_areas)
      surface.associated_thermal_boundaries = [_thermal_boundary]
      _thermal_boundaries.append(_thermal_boundary)
    if self.thermal_archetype is None:
      return None  # there are no archetype
    _number_of_storeys = int(self.volume / self.area / self.thermal_archetype.average_storey_height)
    if _number_of_storeys == 0:
      _number_of_storeys = 1
    _thermal_zone = ThermalZone(_thermal_boundaries, self, self.volume, self.area, _number_of_storeys)
    for thermal_boundary in _thermal_zone.thermal_boundaries:
      thermal_boundary.thermal_zones = [_thermal_zone]
    self._thermal_zones_from_internal_zones = [_thermal_zone]
    return self._thermal_zones_from_internal_zones

  @thermal_zones_from_internal_zones.setter
  def thermal_zones_from_internal_zones(self, value):
    """
    Set city object thermal zones as one per internal zone
    :param value: [ThermalZone]
    """
    self._thermal_zones_from_internal_zones = value



"""
Energy System catalog distribution system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from typing import Union, List, TypeVar

from hub.catalog_factories.data_models.energy_systems.energy_storage_system import EnergyStorageSystem
from hub.catalog_factories.data_models.energy_systems.emission_system import EmissionSystem

GenerationSystem = TypeVar('GenerationSystem')


class DistributionSystem:
  """
  Distribution system class
  """

  def __init__(self, system_id, model_name=None, system_type=None, supply_temperature=None,
               distribution_consumption_fix_flow=None, distribution_consumption_variable_flow=None, heat_losses=None,
               generation_systems=None, energy_storage_systems=None, emission_systems=None):
    self._system_id = system_id
    self._model_name = model_name
    self._type = system_type
    self._supply_temperature = supply_temperature
    self._distribution_consumption_fix_flow = distribution_consumption_fix_flow
    self._distribution_consumption_variable_flow = distribution_consumption_variable_flow
    self._heat_losses = heat_losses
    self._generation_systems = generation_systems
    self._energy_storage_systems = energy_storage_systems
    self._emission_systems = emission_systems

  @property
  def id(self):
    """
    Get system id
    :return: float
    """
    return self._system_id

  @property
  def model_name(self):
    """
    Get model name
    :return: string
    """
    return self._model_name

  @property
  def type(self):
    """
    Get type from [air, water, refrigerant]
    :return: string
    """
    return self._type

  @property
  def supply_temperature(self):
    """
    Get supply_temperature in degree Celsius
    :return: float
    """
    return self._supply_temperature

  @property
  def distribution_consumption_fix_flow(self):
    """
    Get distribution_consumption if the pump or fan work at fix mass or volume flow in ratio over peak power (W/W)
    :return: float
    """
    return self._distribution_consumption_fix_flow

  @property
  def distribution_consumption_variable_flow(self):
    """
    Get distribution_consumption if the pump or fan work at variable mass or volume flow in ratio
    over energy produced (J/J)
    :return: float
    """
    return self._distribution_consumption_variable_flow

  @property
  def heat_losses(self):
    """
    Get heat_losses in ratio over energy produced in J/J
    :return: float
    """
    return self._heat_losses

  @property
  def generation_systems(self) -> Union[None, List[GenerationSystem]]:
    """
    Get generation systems connected to the distribution system
    :return: [GenerationSystem]
    """
    return self._generation_systems

  @property
  def energy_storage_systems(self) -> Union[None, List[EnergyStorageSystem]]:
    """
    Get energy storage systems connected to this distribution system
    :return: [EnergyStorageSystem]
    """
    return self._energy_storage_systems

  @property
  def emission_systems(self) -> Union[None, List[EmissionSystem]]:
    """
    Get energy emission systems connected to this distribution system
    :return: [EmissionSystem]
    """
    return self._emission_systems

  def to_dictionary(self):
    """Class content to dictionary"""
    _generation_systems = [_generation_system.to_dictionary() for _generation_system in
                           self.generation_systems] if self.generation_systems is not None else None
    _energy_storage_systems = [_energy_storage_system.to_dictionary() for _energy_storage_system in
                               self.energy_storage_systems] if self.energy_storage_systems is not None else None
    _emission_systems = [_emission_system.to_dictionary() for _emission_system in
                         self.emission_systems] if self.emission_systems is not None else None

    content = {
      'Layer': {
        'id': self.id,
        'model name': self.model_name,
        'type': self.type,
        'supply temperature [Celsius]': self.supply_temperature,
        'distribution consumption if fix flow over peak power [W/W]': self.distribution_consumption_fix_flow,
        'distribution consumption if variable flow over peak power [J/J]': self.distribution_consumption_variable_flow,
        'heat losses per energy produced [J/J]': self.heat_losses,
        'generation systems connected': _generation_systems,
        'energy storage systems connected': _energy_storage_systems,
        'emission systems connected': _emission_systems
      }
    }
    return content



"""
Montreal custom energy systems catalog module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from ast import literal_eval
import xmltodict
from hub.catalog_factories.catalog import Catalog
from hub.catalog_factories.data_models.energy_systems.system import System
from hub.catalog_factories.data_models.energy_systems.content import Content
from hub.catalog_factories.data_models.energy_systems.non_pv_generation_system import NonPvGenerationSystem
from hub.catalog_factories.data_models.energy_systems.pv_generation_system import PvGenerationSystem
from hub.catalog_factories.data_models.energy_systems.distribution_system import DistributionSystem
from hub.catalog_factories.data_models.energy_systems.emission_system import EmissionSystem
from hub.catalog_factories.data_models.energy_systems.archetype import Archetype
from hub.catalog_factories.data_models.energy_systems.thermal_storage_system import ThermalStorageSystem
from hub.catalog_factories.data_models.energy_systems.electrical_storage_system import ElectricalStorageSystem


class MontrealCustomCatalog(Catalog):
  """
  Montreal custom energy systems catalog class
  """

  def __init__(self, path):
    path = str(path / 'montreal_custom_systems.xml')
    with open(path, 'r', encoding='utf-8') as xml:
      self._archetypes = xmltodict.parse(xml.read(), force_list=('system', 'system_cluster', 'equipment',
                                                                 'demand', 'system_id'))

    self._catalog_generation_equipments = self._load_generation_equipments()
    self._catalog_emission_equipments = self._load_emission_equipments()
    self._catalog_distribution_equipments = self._load_distribution_equipments()
    self._catalog_systems = self._load_systems()
    self._catalog_archetypes = self._load_archetypes()
    # store the full catalog data model in self._content
    self._content = Content(self._catalog_archetypes,
                            self._catalog_systems,
                            self._catalog_generation_equipments,
                            self._catalog_distribution_equipments)

  def _load_generation_equipments(self):
    _equipments = []
    _storages = []
    equipments = self._archetypes['catalog']['generation_equipments']['equipment']
    for equipment in equipments:
      equipment_id = float(equipment['@id'])
      equipment_type = equipment['@type']
      fuel_type = equipment['@fuel_type']
      model_name = equipment['name']
      heating_efficiency = None
      if 'heating_efficiency' in equipment:
        heating_efficiency = float(equipment['heating_efficiency'])
      cooling_efficiency = None
      if 'cooling_efficiency' in equipment:
        cooling_efficiency = float(equipment['cooling_efficiency'])
      electricity_efficiency = None
      if 'electrical_efficiency' in equipment:
        electricity_efficiency = float(equipment['electrical_efficiency'])

      storage_systems = None
      storage = literal_eval(equipment['storage'].capitalize())
      if storage:
        if equipment_type == 'electricity generator':
          storage_system = ElectricalStorageSystem(equipment_id)
        else:
          storage_system = ThermalStorageSystem(equipment_id)
        storage_systems = [storage_system]
      if model_name == 'PV system':
        system_type = 'photovoltaic'
        generation_system = PvGenerationSystem(equipment_id,
                                               name=None,
                                               system_type=system_type,
                                               model_name=model_name,
                                               electricity_efficiency=electricity_efficiency,
                                               energy_storage_systems=storage_systems
                                               )
      else:
        generation_system = NonPvGenerationSystem(equipment_id,
                                                  name=None,
                                                  model_name=model_name,
                                                  system_type=equipment_type,
                                                  fuel_type=fuel_type,
                                                  heat_efficiency=heating_efficiency,
                                                  cooling_efficiency=cooling_efficiency,
                                                  electricity_efficiency=electricity_efficiency,
                                                  energy_storage_systems=storage_systems,
                                                  domestic_hot_water=False
                                                  )
      _equipments.append(generation_system)

    return _equipments

  def _load_distribution_equipments(self):
    _equipments = []
    equipments = self._archetypes['catalog']['distribution_equipments']['equipment']
    for equipment in equipments:
      equipment_id = float(equipment['@id'])
      equipment_type = equipment['@type']
      model_name = equipment['name']
      distribution_heat_losses = None
      if 'distribution_heat_losses' in equipment:
        distribution_heat_losses = float(equipment['distribution_heat_losses']['#text']) / 100
      distribution_consumption_fix_flow = None
      if 'distribution_consumption_fix_flow' in equipment:
        distribution_consumption_fix_flow = float(equipment['distribution_consumption_fix_flow']['#text']) / 100
      distribution_consumption_variable_flow = None
      if 'distribution_consumption_variable_flow' in equipment:
        distribution_consumption_variable_flow = float(
          equipment['distribution_consumption_variable_flow']['#text']) / 100

      emission_equipment = equipment['dissipation_id']
      _emission_equipments = None
      for equipment_archetype in self._catalog_emission_equipments:
        if int(equipment_archetype.id) == int(emission_equipment):
          _emission_equipments = [equipment_archetype]

      distribution_system = DistributionSystem(equipment_id,
                                               model_name=model_name,
                                               system_type=equipment_type,
                                               distribution_consumption_fix_flow=distribution_consumption_fix_flow,
                                               distribution_consumption_variable_flow=distribution_consumption_variable_flow,
                                               heat_losses=distribution_heat_losses,
                                               emission_systems=_emission_equipments)

      _equipments.append(distribution_system)
    return _equipments

  def _load_emission_equipments(self):
    _equipments = []
    equipments = self._archetypes['catalog']['dissipation_equipments']['equipment']
    for equipment in equipments:
      equipment_id = float(equipment['@id'])
      equipment_type = equipment['@type']
      model_name = equipment['name']
      parasitic_consumption = 0
      if 'parasitic_consumption' in equipment:
        parasitic_consumption = float(equipment['parasitic_consumption']['#text']) / 100

      emission_system = EmissionSystem(equipment_id,
                                       model_name=model_name,
                                       system_type=equipment_type,
                                       parasitic_energy_consumption=parasitic_consumption)

      _equipments.append(emission_system)
    return _equipments

  def _load_systems(self):
    _catalog_systems = []
    systems = self._archetypes['catalog']['systems']['system']
    for system in systems:
      system_id = float(system['@id'])
      name = system['name']
      demands = system['demands']['demand']
      generation_equipment = system['equipments']['generation_id']
      _generation_equipments = None
      for equipment_archetype in self._catalog_generation_equipments:
        if int(equipment_archetype.id) == int(generation_equipment):
          _generation_equipments = [equipment_archetype]
      distribution_equipment = system['equipments']['distribution_id']
      _distribution_equipments = None
      for equipment_archetype in self._catalog_distribution_equipments:
        if int(equipment_archetype.id) == int(distribution_equipment):
          _distribution_equipments = [equipment_archetype]

      _catalog_systems.append(System(system_id,
                                     demands,
                                     name=name,
                                     generation_systems=_generation_equipments,
                                     distribution_systems=_distribution_equipments))
    return _catalog_systems

  def _load_archetypes(self):
    _catalog_archetypes = []
    system_clusters = self._archetypes['catalog']['system_clusters']['system_cluster']
    for system_cluster in system_clusters:
      name = system_cluster['@name']
      systems = system_cluster['systems']['system_id']
      _systems = []
      for system in systems:
        for system_archetype in self._catalog_systems:
          if int(system_archetype.id) == int(system):
            _systems.append(system_archetype)
      _catalog_archetypes.append(Archetype(name, _systems))
    return _catalog_archetypes

  def names(self, category=None):
    """
    Get the catalog elements names
    :parm: optional category filter
    """
    if category is None:
      _names = {'archetypes': [], 'systems': [], 'generation_equipments': [], 'distribution_equipments': [],
                'emission_equipments': []}
      for archetype in self._content.archetypes:
        _names['archetypes'].append(archetype.name)
      for system in self._content.systems:
        _names['systems'].append(system.name)
      for equipment in self._content.generation_equipments:
        _names['generation_equipments'].append(equipment.model_name)
      for equipment in self._content.distribution_equipments:
        _names['distribution_equipments'].append(equipment.model_name)
    else:
      _names = {category: []}
      if category.lower() == 'archetypes':
        for archetype in self._content.archetypes:
          _names[category].append(archetype.name)
      elif category.lower() == 'systems':
        for system in self._content.systems:
          _names[category].append(system.name)
      elif category.lower() == 'generation_equipments':
        for system in self._content.generation_equipments:
          _names[category].append(system.model_name)
      elif category.lower() == 'distribution_equipments':
        for system in self._content.distribution_equipments:
          _names[category].append(system.model_name)
      else:
        raise ValueError(f'Unknown category [{category}]')
    return _names

  def entries(self, category=None):
    """
    Get the catalog elements
    :parm: optional category filter
    """
    if category is None:
      return self._content
    if category.lower() == 'archetypes':
      return self._content.archetypes
    if category.lower() == 'systems':
      return self._content.systems
    if category.lower() == 'generation_equipments':
      return self._content.generation_equipments
    if category.lower() == 'distribution_equipments':
      return self._content.distribution_equipments

  def get_entry(self, name):
    """
    Get one catalog element by names
    :parm: entry name
    """
    for entry in self._content.archetypes:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.systems:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.generation_equipments:
      if entry.model_name.lower() == name.lower():
        return entry
    for entry in self._content.distribution_equipments:
      if entry.model_name.lower() == name.lower():
        return entry
    raise IndexError(f"{name} doesn't exists in the catalog")



"""
Dictionaries module for Pluto function to hub function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import hub.helpers.constants as cte


class PlutoFunctionToHubFunction:
  """
  Pluto function to hub function class
  """
  def __init__(self):
    self._dictionary = {
      'A0': cte.SINGLE_FAMILY_HOUSE,
      'A1': cte.SINGLE_FAMILY_HOUSE,
      'A2': cte.SINGLE_FAMILY_HOUSE,
      'A3': cte.SINGLE_FAMILY_HOUSE,
      'A4': cte.SINGLE_FAMILY_HOUSE,
      'A5': cte.SINGLE_FAMILY_HOUSE,
      'A6': cte.SINGLE_FAMILY_HOUSE,
      'A7': cte.SINGLE_FAMILY_HOUSE,
      'A8': cte.SINGLE_FAMILY_HOUSE,
      'A9': cte.SINGLE_FAMILY_HOUSE,
      'B1': cte.MULTI_FAMILY_HOUSE,
      'B2': cte.MULTI_FAMILY_HOUSE,
      'B3': cte.MULTI_FAMILY_HOUSE,
      'B9': cte.MULTI_FAMILY_HOUSE,
      'C0': cte.RESIDENTIAL,
      'C1': cte.RESIDENTIAL,
      'C2': cte.RESIDENTIAL,
      'C3': cte.RESIDENTIAL,
      'C4': cte.RESIDENTIAL,
      'C5': cte.RESIDENTIAL,
      'C6': cte.RESIDENTIAL,
      'C7': cte.RESIDENTIAL,
      'C8': cte.RESIDENTIAL,
      'C9': cte.RESIDENTIAL,
      'D0': cte.RESIDENTIAL,
      'D1': cte.RESIDENTIAL,
      'D2': cte.RESIDENTIAL,
      'D3': cte.RESIDENTIAL,
      'D4': cte.RESIDENTIAL,
      'D5': cte.RESIDENTIAL,
      'D6': cte.RESIDENTIAL,
      'D7': cte.RESIDENTIAL,
      'D8': cte.RESIDENTIAL,
      'D9': cte.RESIDENTIAL,
      'E1': cte.WAREHOUSE,
      'E3': cte.WAREHOUSE,
      'E4': cte.WAREHOUSE,
      'E5': cte.WAREHOUSE,
      'E7': cte.WAREHOUSE,
      'E9': cte.WAREHOUSE,
      'F1': cte.WAREHOUSE,
      'F2': cte.WAREHOUSE,
      'F4': cte.WAREHOUSE,
      'F5': cte.WAREHOUSE,
      'F8': cte.WAREHOUSE,
      'F9': cte.WAREHOUSE,
      'G0': cte.SMALL_OFFICE,
      'G1': cte.SMALL_OFFICE,
      'G2': cte.SMALL_OFFICE,
      'G3': cte.SMALL_OFFICE,
      'G4': cte.SMALL_OFFICE,
      'G5': cte.SMALL_OFFICE,
      'G6': cte.SMALL_OFFICE,
      'G7': cte.SMALL_OFFICE,
      'G8': cte.SMALL_OFFICE,
      'G9': cte.SMALL_OFFICE,
      'H1': cte.HOTEL,
      'H2': cte.HOTEL,
      'H3': cte.HOTEL,
      'H4': cte.HOTEL,
      'H5': cte.HOTEL,
      'H6': cte.HOTEL,
      'H7': cte.HOTEL,
      'H8': cte.HOTEL,
      'H9': cte.HOTEL,
      'HB': cte.HOTEL,
      'HH': cte.HOTEL,
      'HR': cte.HOTEL,
      'HS': cte.HOTEL,
      'I1': cte.HOSPITAL,
      'I2': cte.OUT_PATIENT_HEALTH_CARE,
      'I3': cte.OUT_PATIENT_HEALTH_CARE,
      'I4': cte.RESIDENTIAL,
      'I5': cte.OUT_PATIENT_HEALTH_CARE,
      'I6': cte.OUT_PATIENT_HEALTH_CARE,
      'I7': cte.OUT_PATIENT_HEALTH_CARE,
      'I9': cte.OUT_PATIENT_HEALTH_CARE,
      'J1': cte.LARGE_OFFICE,
      'J2': cte.LARGE_OFFICE,
      'J3': cte.LARGE_OFFICE,
      'J4': cte.LARGE_OFFICE,
      'J5': cte.LARGE_OFFICE,
      'J6': cte.LARGE_OFFICE,
      'J7': cte.LARGE_OFFICE,
      'J8': cte.LARGE_OFFICE,
      'J9': cte.LARGE_OFFICE,
      'K1': cte.STRIP_MALL,
      'K2': cte.STRIP_MALL,
      'K3': cte.STRIP_MALL,
      'K4': cte.RESIDENTIAL,
      'K5': cte.RESTAURANT,
      'K6': cte.SUPERMARKET,
      'K7': cte.SUPERMARKET,
      'K8': cte.SUPERMARKET,
      'K9': cte.SUPERMARKET,
      'L1': cte.RESIDENTIAL,
      'L2': cte.RESIDENTIAL,
      'L3': cte.RESIDENTIAL,
      'L8': cte.RESIDENTIAL,
      'L9': cte.RESIDENTIAL,
      'M1': cte.LARGE_OFFICE,
      'M2': cte.LARGE_OFFICE,
      'M3': cte.LARGE_OFFICE,
      'M4': cte.LARGE_OFFICE,
      'M9': cte.LARGE_OFFICE,
      'N1': cte.RESIDENTIAL,
      'N2': cte.RESIDENTIAL,
      'N3': cte.RESIDENTIAL,
      'N4': cte.RESIDENTIAL,
      'N9': cte.RESIDENTIAL,
      'O1': cte.SMALL_OFFICE,
      'O2': cte.SMALL_OFFICE,
      'O3': cte.SMALL_OFFICE,
      'O4': cte.SMALL_OFFICE,
      'O5': cte.SMALL_OFFICE,
      'O6': cte.SMALL_OFFICE,
      'O7': cte.SMALL_OFFICE,
      'O8': cte.SMALL_OFFICE,
      'O9': cte.SMALL_OFFICE,
      'P1': cte.LARGE_OFFICE,
      'P2': cte.HOTEL,
      'P3': cte.SMALL_OFFICE,
      'P4': cte.SMALL_OFFICE,
      'P5': cte.SMALL_OFFICE,
      'P6': cte.SMALL_OFFICE,
      'P7': cte.LARGE_OFFICE,
      'P8': cte.LARGE_OFFICE,
      'P9': cte.SMALL_OFFICE,
      'Q0': cte.SMALL_OFFICE,
      'Q1': cte.SMALL_OFFICE,
      'Q2': cte.SMALL_OFFICE,
      'Q3': cte.SMALL_OFFICE,
      'Q4': cte.SMALL_OFFICE,
      'Q5': cte.SMALL_OFFICE,
      'Q6': cte.SMALL_OFFICE,
      'Q7': cte.SMALL_OFFICE,
      'Q8': cte.SMALL_OFFICE,
      'Q9': cte.SMALL_OFFICE,
      'R0': cte.RESIDENTIAL,
      'R1': cte.RESIDENTIAL,
      'R2': cte.RESIDENTIAL,
      'R3': cte.RESIDENTIAL,
      'R4': cte.RESIDENTIAL,
      'R5': cte.RESIDENTIAL,
      'R6': cte.RESIDENTIAL,
      'R7': cte.RESIDENTIAL,
      'R8': cte.RESIDENTIAL,
      'R9': cte.RESIDENTIAL,
      'RA': cte.RESIDENTIAL,
      'RB': cte.RESIDENTIAL,
      'RC': cte.RESIDENTIAL,
      'RD': cte.RESIDENTIAL,
      'RG': cte.RESIDENTIAL,
      'RH': cte.RESIDENTIAL,
      'RI': cte.RESIDENTIAL,
      'RK': cte.RESIDENTIAL,
      'RM': cte.RESIDENTIAL,
      'RR': cte.RESIDENTIAL,
      'RS': cte.RESIDENTIAL,
      'RW': cte.RESIDENTIAL,
      'RX': cte.RESIDENTIAL,
      'RZ': cte.RESIDENTIAL,
      'S0': cte.RESIDENTIAL,
      'S1': cte.RESIDENTIAL,
      'S2': cte.RESIDENTIAL,
      'S3': cte.RESIDENTIAL,
      'S4': cte.RESIDENTIAL,
      'S5': cte.RESIDENTIAL,
      'S9': cte.RESIDENTIAL,
      'U0': cte.WAREHOUSE,
      'U1': cte.WAREHOUSE,
      'U2': cte.WAREHOUSE,
      'U3': cte.WAREHOUSE,
      'U4': cte.WAREHOUSE,
      'U5': cte.WAREHOUSE,
      'U6': cte.WAREHOUSE,
      'U7': cte.WAREHOUSE,
      'U8': cte.WAREHOUSE,
      'U9': cte.WAREHOUSE,
      'W1': cte.PRIMARY_SCHOOL,
      'W2': cte.PRIMARY_SCHOOL,
      'W3': cte.SECONDARY_SCHOOL,
      'W4': cte.EDUCATION,
      'W5': cte.SECONDARY_SCHOOL,
      'W6': cte.SECONDARY_SCHOOL,
      'W7': cte.SECONDARY_SCHOOL,
      'W8': cte.PRIMARY_SCHOOL,
      'W9': cte.SECONDARY_SCHOOL,
      'Y1': cte.LARGE_OFFICE,
      'Y2': cte.LARGE_OFFICE,
      'Y3': cte.LARGE_OFFICE,
      'Y4': cte.LARGE_OFFICE,
      'Y5': cte.LARGE_OFFICE,
      'Y6': cte.LARGE_OFFICE,
      'Y7': cte.LARGE_OFFICE,
      'Y8': cte.LARGE_OFFICE,
      'Y9': cte.LARGE_OFFICE,
      'Z1': cte.LARGE_OFFICE
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Energy System catalog thermal storage system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from hub.catalog_factories.data_models.energy_systems.energy_storage_system import EnergyStorageSystem
from hub.catalog_factories.data_models.construction.layer import Layer
from hub.catalog_factories.data_models.construction.material import Material

class ThermalStorageSystem(EnergyStorageSystem):
  """"
  Energy Storage System Class
  """

  def __init__(self, storage_id, type_energy_stored=None, model_name=None, manufacturer=None, storage_type=None,
               nominal_capacity=None, losses_ratio=None, volume=None, height=None, layers=None,
               maximum_operating_temperature=None, storage_medium=None, heating_coil_capacity=None):

    super().__init__(storage_id, model_name, manufacturer, nominal_capacity, losses_ratio)
    self._type_energy_stored = type_energy_stored
    self._storage_type = storage_type
    self._volume = volume
    self._height = height
    self._layers = layers
    self._maximum_operating_temperature = maximum_operating_temperature
    self._storage_medium = storage_medium
    self._heating_coil_capacity = heating_coil_capacity

  @property
  def type_energy_stored(self):
    """
    Get type of energy stored from ['electrical', 'thermal']
    :return: string
    """
    return self._type_energy_stored

  @property
  def storage_type(self):
    """
    Get storage type from ['thermal', 'sensible', 'latent']
    :return: string
    """
    return self._storage_type

  @property
  def volume(self):
    """
    Get the physical volume of the storage system in cubic meters
    :return: float
    """
    return self._volume

  @property
  def height(self):
    """
    Get the diameter of the storage system in meters
    :return: float
    """
    return self._height

  @property
  def layers(self) -> [Layer]:
    """
    Get construction layers
    :return: [layer]
    """
    return self._layers

  @property
  def maximum_operating_temperature(self):
    """
    Get maximum operating temperature of the storage system in degree Celsius
    :return: float
    """
    return self._maximum_operating_temperature

  @property
  def storage_medium(self) -> Material:
    """
    Get thermodynamic characteristics of the storage medium
    :return: [material
    """
    return self._storage_medium

  @property
  def heating_coil_capacity(self):
    """
    Get heating coil capacity in Watts
    :return: [material
    """
    return self._heating_coil_capacity

  def to_dictionary(self):
    """Class content to dictionary"""
    _layers = None
    _medias = None
    if self.layers is not None:
      _layers = []
      for _layer in self.layers:
        _layers.append(_layer.to_dictionary())

    if self.storage_medium is not None:
      _medias = self.storage_medium.to_dictionary()

    content = {
      'Storage component':
      {
        'storage id': self.id,
        'type of energy stored': self.type_energy_stored,
        'model name': self.model_name,
        'manufacturer': self.manufacturer,
        'storage type': self.storage_type,
        'nominal capacity [J]': self.nominal_capacity,
        'losses-ratio [J/J]': self.losses_ratio,
        'volume [m3]': self.volume,
        'height [m]': self.height,
        'layers': _layers,
        'maximum operating temperature [Celsius]': self.maximum_operating_temperature,
        'storage_medium': _medias,
        'heating coil capacity [W]': self.heating_coil_capacity
      }
    }
    return content



"""
Energy System catalog electrical storage system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from hub.catalog_factories.data_models.energy_systems.energy_storage_system import EnergyStorageSystem


class ElectricalStorageSystem(EnergyStorageSystem):
  """"
  Energy Storage System Class
  """

  def __init__(self, storage_id, type_energy_stored=None, model_name=None, manufacturer=None, storage_type=None,
               nominal_capacity=None, losses_ratio=None, rated_output_power=None, nominal_efficiency=None,
               battery_voltage=None, depth_of_discharge=None, self_discharge_rate=None):

    super().__init__(storage_id, model_name, manufacturer, nominal_capacity, losses_ratio)
    self._type_energy_stored = type_energy_stored
    self._storage_type = storage_type
    self._rated_output_power = rated_output_power
    self._nominal_efficiency = nominal_efficiency
    self._battery_voltage = battery_voltage
    self._depth_of_discharge = depth_of_discharge
    self._self_discharge_rate = self_discharge_rate

  @property
  def type_energy_stored(self):
    """
    Get type of energy stored from ['electrical', 'thermal']
    :return: string
    """
    return self._type_energy_stored

  @property
  def storage_type(self):
    """
    Get storage type from ['lithium_ion', 'lead_acid', 'NiCd']
    :return: string
    """
    return self._storage_type

  @property
  def rated_output_power(self):
    """
    Get the rated output power of storage system in Watts
    :return: float
    """
    return self._rated_output_power

  @property
  def nominal_efficiency(self):
    """
     Get the nominal efficiency of the storage system
    :return: float
    """
    return self._nominal_efficiency

  @property
  def battery_voltage(self):
    """
    Get the battery voltage in Volts
    :return: float
    """
    return self._battery_voltage

  @property
  def depth_of_discharge(self):
    """
    Get the depth of discharge as a percentage
    :return: float
    """
    return self._depth_of_discharge

  @property
  def self_discharge_rate(self):
    """
    Get the self discharge rate of battery as a percentage
    :return: float
    """
    return self._self_discharge_rate

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Storage component': {
      'storage id': self.id,
      'type of energy stored': self.type_energy_stored,
      'model name': self.model_name,
      'manufacturer': self.manufacturer,
      'storage type': self.storage_type,
      'nominal capacity [J]': self.nominal_capacity,
      'losses-ratio [J/J]': self.losses_ratio,
      'rated power [W]': self.rated_output_power,
      'nominal efficiency': self.nominal_efficiency,
      'battery voltage [V]': self.battery_voltage,
      'depth of discharge [%]': self.depth_of_discharge,
      'self discharge rate': self.self_discharge_rate
    }
    }
    return content



"""
TestExports test and validate the city export formats
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import subprocess
from pathlib import Path
from unittest import TestCase

import hub.helpers.constants as cte
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.exports.exports_factory import ExportsFactory
from hub.helpers.dictionaries import Dictionaries
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.results_factory import ResultFactory
from hub.imports.usage_factory import UsageFactory


class TestResultsImport(TestCase):
  """
  TestImports class contains the unittest for import functionality
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()
    file = 'Citylayers_neighbours_simp2.json'
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('geojson',
                                 path=file_path,
                                 height_field='heightmax',
                                 year_of_construction_field='ANNEE_CONS',
                                 function_field='CODE_UTILI',
                                 function_to_hub=Dictionaries().montreal_function_to_hub_function).city

    ConstructionFactory('nrcan', self._city).enrich()
    UsageFactory('comnet', self._city).enrich()

  def test_sra_import(self):
    ExportsFactory('sra', self._city, self._output_path).export()
    sra_path = (self._output_path / f'{self._city.name}_sra.xml').resolve()
    subprocess.run(['sra', str(sra_path)])
    ResultFactory('sra', self._city, self._output_path).enrich()
    # Check that all the buildings have radiance in the surfaces
    for building in self._city.buildings:
      for surface in building.surfaces:
        self.assertIsNotNone(surface.global_irradiance)

  def test_meb_import(self):
    ExportsFactory('sra', self._city, self._output_path).export()
    sra_path = (self._output_path / f'{self._city.name}_sra.xml').resolve()
    subprocess.run(['sra', str(sra_path)])
    ResultFactory('sra', self._city, self._output_path).enrich()
    EnergyBuildingsExportsFactory('insel_monthly_energy_balance', self._city, self._output_path).export()
    for building in self._city.buildings:
      insel_path = (self._output_path / f'{building.name}.insel')
      subprocess.run(['insel', str(insel_path)])
    ResultFactory('insel_monthly_energy_balance', self._city, self._output_path).enrich()
    # Check that all the buildings have heating and cooling values
    for building in self._city.buildings:
      self.assertIsNotNone(building.heating_demand[cte.MONTH])
      self.assertIsNotNone(building.cooling_demand[cte.MONTH])
      self.assertIsNotNone(building.heating_demand[cte.YEAR])
      self.assertIsNotNone(building.cooling_demand[cte.YEAR])
      self.assertIsNotNone(building.lighting_peak_load[cte.MONTH])
      self.assertIsNotNone(building.lighting_peak_load[cte.YEAR])
      self.assertIsNotNone(building.appliances_peak_load[cte.MONTH])
      self.assertIsNotNone(building.appliances_peak_load[cte.YEAR])

  def test_peak_loads(self):
    # todo: this is not technically a import
    ExportsFactory('sra', self._city, self._output_path).export()
    sra_path = (self._output_path / f'{self._city.name}_sra.xml').resolve()
    subprocess.run(['sra', str(sra_path)])
    ResultFactory('sra', self._city, self._output_path).enrich()
    for building in self._city.buildings:
      self.assertIsNotNone(building.heating_peak_load)
      self.assertIsNotNone(building.cooling_peak_load)

    values = [0 for _ in range(8760)]
    values[0] = 1000
    expected_monthly_list = [0 for _ in range(12)]
    expected_monthly_list[0] = 1000
    for building in self._city.buildings:
      building.heating_demand[cte.HOUR] = values
      building.cooling_demand[cte.HOUR] = values
      self.assertIsNotNone(building.heating_peak_load)
      self.assertIsNotNone(building.cooling_peak_load)

  def test_energy_plus_results_import(self):
    ResultFactory('energy_plus_single_building', self._city, self._example_path).enrich()
    for building in self._city.buildings:
      csv_output_name = f'{building.name}_out.csv'
      csv_output_path = (self._example_path / csv_output_name).resolve()
      if csv_output_path.is_file():
        self.assertEqual(building.name, '12')
        self.assertIsNotNone(building.heating_demand)
        self.assertIsNotNone(building.cooling_demand)
        self.assertIsNotNone(building.domestic_hot_water_heat_demand)
        self.assertIsNotNone(building.lighting_electrical_demand)
        self.assertIsNotNone(building.appliances_electrical_demand)
        total_demand = sum(building.heating_demand[cte.HOUR])
        self.assertAlmostEqual(total_demand, building.heating_demand[cte.YEAR][0], 3)
        total_demand = sum(building.heating_demand[cte.MONTH])
        self.assertEqual(total_demand, building.heating_demand[cte.YEAR][0], 3)
      if building.name != '12':
        self.assertDictEqual(building.heating_demand, {})
        self.assertDictEqual(building.cooling_demand, {})
        self.assertDictEqual(building.domestic_hot_water_heat_demand, {})
        self.assertDictEqual(building.lighting_electrical_demand, {})
        self.assertDictEqual(building.appliances_electrical_demand, {})

  def test_energy_plus_multiple_buildings_results_import(self):
    ResultFactory('energy_plus_multiple_buildings', self._city, self._example_path).enrich()
    csv_output_name = f'{self._city.name}_out.csv'
    csv_output_path = (self._example_path / csv_output_name).resolve()
    if csv_output_path.is_file():
      for building in self._city.buildings:
        self.assertIsNotNone(building.heating_demand)
        self.assertIsNotNone(building.cooling_demand)
        self.assertIsNotNone(building.domestic_hot_water_heat_demand)
        self.assertIsNotNone(building.lighting_electrical_demand)
        self.assertIsNotNone(building.appliances_electrical_demand)
        total_demand = sum(building.heating_demand[cte.HOUR])
        self.assertAlmostEqual(total_demand, building.heating_demand[cte.YEAR][0], 2)
        total_demand = sum(building.heating_demand[cte.MONTH])
        self.assertEqual(total_demand, building.heating_demand[cte.YEAR][0], 2)



"""
Record module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class Record:
  """
  Record class
  """

  def __init__(self, time=None, value=None, flag=None):
    self._time = time
    self._value = value
    self._flag = flag

  @property
  def time(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._time

  @property
  def value(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._value

  @property
  def flag(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._flag



"""
TestUsageCatalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from unittest import TestCase
from hub.catalog_factories.usage_catalog_factory import UsageCatalogFactory


class TestConstructionCatalog(TestCase):
  def test_comnet_catalog(self):
    catalog = UsageCatalogFactory('comnet').catalog
    self.assertIsNotNone(catalog, 'catalog is none')
    content = catalog.entries()
    self.assertEqual(32, len(content.usages), 'Wrong number of usages')

  def test_nrcan_catalog(self):
    catalog = UsageCatalogFactory('nrcan').catalog
    self.assertIsNotNone(catalog, 'catalog is none')
    content = catalog.entries()
    self.assertEqual(34, len(content.usages), 'Wrong number of usages')

  def test_palma_catalog(self):
    catalog = UsageCatalogFactory('palma').catalog
    self.assertIsNotNone(catalog, 'catalog is none')
    content = catalog.entries()
    #print(catalog.entries())
    self.assertEqual(1, len(content.usages), 'Wrong number of usages')



"""
Energy System catalog heat generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from __future__ import annotations
from abc import ABC
from typing import List, Union

from hub.catalog_factories.data_models.energy_systems.energy_storage_system import EnergyStorageSystem
from hub.catalog_factories.data_models.energy_systems.distribution_system import DistributionSystem


class GenerationSystem(ABC):
  """
  Heat Generation system class
  """

  def __init__(self, system_id, name, model_name=None, manufacturer=None, fuel_type=None,
               distribution_systems=None, energy_storage_systems=None):
    self._system_id = system_id
    self._name = name
    self._model_name = model_name
    self._manufacturer = manufacturer
    self._fuel_type = fuel_type
    self._distribution_systems = distribution_systems
    self._energy_storage_systems = energy_storage_systems

  @property
  def id(self):
    """
    Get system id
    :return: float
    """
    return self._system_id

  @property
  def name(self):
    """
    Get system name
    :return: string
    """
    return self._name

  @property
  def system_type(self):
    """
    Get type
    :return: string
    """
    raise NotImplementedError

  @property
  def model_name(self):
    """
    Get system id
    :return: float
    """
    return self._model_name

  @property
  def manufacturer(self):
    """
    Get name
    :return: string
    """
    return self._manufacturer

  @property
  def fuel_type(self):
    """
    Get fuel_type from [renewable, gas, diesel, electricity, wood, coal, biogas]
    :return: string
    """
    return self._fuel_type

  @property
  def distribution_systems(self) -> Union[None, List[DistributionSystem]]:
    """
    Get distributions systems connected to this generation system
    :return: [DistributionSystem]
    """
    return self._distribution_systems

  @property
  def energy_storage_systems(self) -> Union[None, List[EnergyStorageSystem]]:
    """
    Get energy storage systems connected to this generation system
    :return: [EnergyStorageSystem]
    """
    return self._energy_storage_systems

  def to_dictionary(self):
    """Class content to dictionary"""
    raise NotImplementedError



"""
Energy System catalog heat generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class NorthAmericaSystemToHubEnergyGenerationSystem:
  """
  Montreal's system to hub energy generation system class
  """
  def __init__(self):
    self._dictionary = {
      'template Natural-Gas Boiler': cte.BOILER,
      'template Electric Boiler': cte.BOILER,
      'template Air-to-Water Heat Pump': cte.HEAT_PUMP,
      'template Groundwater-to-Water Heat Pump': cte.HEAT_PUMP,
      'template Water-to-Water Heat Pump': cte.HEAT_PUMP,
      'template Photovoltaic Module': cte.PHOTOVOLTAIC,
      'e': cte.HEATING,
                        }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Model representation of a City
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

import datetime

from sqlalchemy import Column, Integer, String, Sequence, ForeignKey
from sqlalchemy import DateTime

from hub.persistence.configuration import Models


class City(Models):
  """A model representation of a city
  """
  __tablename__ = 'city'
  id = Column(Integer, Sequence('city_id_seq'), primary_key=True)
  pickle_path = Column(String, nullable=False)
  name = Column(String, nullable=False)
  scenario = Column(String, nullable=False)
  application_id = Column(Integer, ForeignKey('application.id', ondelete='CASCADE'), nullable=False)
  user_id = Column(Integer, ForeignKey('user.id', ondelete='CASCADE'), nullable=True)
  hub_release = Column(String, nullable=False)
  created = Column(DateTime, default=datetime.datetime.utcnow)
  updated = Column(DateTime, default=datetime.datetime.utcnow)

  def __init__(self, pickle_path, name, scenario, application_id, user_id, hub_release):
    self.pickle_path = str(pickle_path)
    self.name = name
    self.scenario = scenario
    self.application_id = application_id
    self.user_id = user_id
    self.hub_release = hub_release



"""
Emission system module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class EmissionSystem:
  """
  EmissionSystem class
  """
  def __init__(self):
    self._model_name = None
    self._type = None
    self._parasitic_energy_consumption = 0

  @property
  def model_name(self):
    """
    Get model name
    :return: string
    """
    return self._model_name

  @model_name.setter
  def model_name(self, value):
    """
    Set model name
    :param value: string
    """
    self._model_name = value

  @property
  def type(self):
    """
    Get type
    :return: string
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set type
    :param value: string
    """
    self._type = value

  @property
  def parasitic_energy_consumption(self):
    """
    Get parasitic_energy_consumption in ratio (W/W)
    :return: float
    """
    return self._parasitic_energy_consumption

  @parasitic_energy_consumption.setter
  def parasitic_energy_consumption(self, value):
    """
    Set parasitic_energy_consumption in ratio (W/W)
    :param value: float
    """
    self._parasitic_energy_consumption = value



"""
export a city into Cesium tileset format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import json
import math

import pyproj
from pyproj import Transformer

from hub.helpers.geometry_helper import GeometryHelper


class CesiumjsTileset:
  def __init__(self, city, file_name, target_buildings=None, base_uri=None):
    self._city = city
    self._file_name = file_name
    self._target_buildings = target_buildings
    if base_uri is None:
      base_uri = '.'
    self._base_uri = base_uri
    try:
      srs_name = self._city.srs_name
      if self._city.srs_name in GeometryHelper.srs_transformations:
        srs_name = GeometryHelper.srs_transformations[self._city.srs_name]
      input_reference = pyproj.CRS(srs_name)  # Projected coordinate system from input data
    except pyproj.exceptions.CRSError as err:
      raise pyproj.exceptions.CRSError from err
    self._to_gps = Transformer.from_crs(input_reference, pyproj.CRS('EPSG:4326'))
    city_upper_corner = [
      self._city.upper_corner[0] - self._city.lower_corner[0],
      self._city.upper_corner[1] - self._city.lower_corner[1],
      self._city.upper_corner[2] - self._city.lower_corner[2]
    ]
    city_lower_corner = [0, 0, 0]
    self._tile_set = {
      'asset': {
        'version': '1.1',
        "tilesetVersion": "1.2.3"
      },
      'position': self._to_gps.transform(self._city.lower_corner[0], self._city.lower_corner[1]),
      'schema': {
        'id': "building",
        'classes': {
          'building': {
            "properties": {
              'name': {
                'type': 'STRING'
              },
              'position': {
                'type': 'SCALAR',
                'array': True,
                'componentType': 'FLOAT32'
              },
              'aliases': {
                'type': 'STRING',
                'array': True,
              },
              'volume': {
                'type': 'SCALAR',
                'componentType': 'FLOAT32'
              },
              'floor_area': {
                'type': 'SCALAR',
                'componentType': 'FLOAT32'
              },
              'max_height': {
                'type': 'SCALAR',
                'componentType': 'INT32'
              },
              'year_of_construction': {
                'type': 'SCALAR',
                'componentType': 'INT32'
              },
              'function': {
                'type': 'STRING'
              },
              'usages_percentage': {
                'type': 'STRING'
              }
            }
          }
        }
      },
      'geometricError': 500,
      'root': {
        'boundingVolume': {
          'box': CesiumjsTileset._box_values(city_upper_corner, city_lower_corner)
        },
        'geometricError': 70,
        'refine': 'ADD',
        'children': []
      }
    }

    self._export()

  @staticmethod
  def _box_values(upper_corner, lower_corner):

    x = (upper_corner[0] - lower_corner[0]) / 2
    x_center = ((upper_corner[0] - lower_corner[0]) / 2) + lower_corner[0]
    y = (upper_corner[1] - lower_corner[1]) / 2
    y_center = ((upper_corner[1] - lower_corner[1]) / 2) + lower_corner[1]
    z = (upper_corner[2] - lower_corner[2]) / 2
    return [x_center, y_center, z, x, 0, 0, 0, y, 0, 0, 0, z]

  def _ground_coordinates(self, coordinates):
    ground_coordinates = []
    for coordinate in coordinates:
      ground_coordinates.append(
        (coordinate[0] - self._city.lower_corner[0], coordinate[1] - self._city.lower_corner[1])
      )
    return ground_coordinates

  def _export(self):
    for building in self._city.buildings:
      upper_corner = [-math.inf, -math.inf, 0]
      lower_corner = [math.inf, math.inf, 0]
      lower_corner_coordinates = lower_corner
      for surface in building.grounds:  # todo: maybe we should add the terrain?
        coordinates = self._ground_coordinates(surface.solid_polygon.coordinates)
        lower_corner = [min([c[0] for c in coordinates]), min([c[1] for c in coordinates]), 0]
        lower_corner_coordinates = [
          min([c[0] for c in surface.solid_polygon.coordinates]),
          min([c[1] for c in surface.solid_polygon.coordinates]),
          0
        ]
        upper_corner = [max([c[0] for c in coordinates]), max([c[1] for c in coordinates]), building.max_height]

      tile = {
        'boundingVolume': {
          'box': CesiumjsTileset._box_values(upper_corner, lower_corner)
        },
        'geometricError': 250,
        'metadata': {
          'class': 'building',
          'properties': {
            'name': building.name,
            'position': self._to_gps.transform(lower_corner_coordinates[0], lower_corner_coordinates[1]),
            'aliases': building.aliases,
            'volume': building.volume,
            'floor_area': building.floor_area,
            'max_height': building.max_height,
            'year_of_construction': building.year_of_construction,
            'function': building.function,
            'usages_percentage': building.usages_percentage
          }
        },
        'content': {
          'uri': f'{self._base_uri}/{building.name}.glb'
        }
      }
      self._tile_set['root']['children'].append(tile)

    with open(self._file_name, 'w') as f:
      json.dump(self._tile_set, f, indent=2)



"""
Montreal custom energy system importer
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Project Contributor Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

import logging
import copy

from hub.catalog_factories.energy_systems_catalog_factory import EnergySystemsCatalogFactory
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.city_model_structure.energy_systems.distribution_system import DistributionSystem
from hub.city_model_structure.energy_systems.non_pv_generation_system import NonPvGenerationSystem
from hub.city_model_structure.energy_systems.pv_generation_system import PvGenerationSystem
from hub.city_model_structure.energy_systems.electrical_storage_system import ElectricalStorageSystem
from hub.city_model_structure.energy_systems.thermal_storage_system import ThermalStorageSystem
from hub.city_model_structure.energy_systems.emission_system import EmissionSystem
from hub.helpers.dictionaries import Dictionaries


class MontrealCustomEnergySystemParameters:
  """
  MontrealCustomEnergySystemParameters class
  """

  def __init__(self, city):
    self._city = city

  def enrich_buildings(self):
    """
    Returns the city with the system parameters assigned to the buildings
    :return:
    """
    city = self._city
    montreal_custom_catalog = EnergySystemsCatalogFactory('montreal_custom').catalog
    if city.generic_energy_systems is None:
      _generic_energy_systems = {}
    else:
      _generic_energy_systems = city.generic_energy_systems
    for building in city.buildings:
      archetype_name = building.energy_systems_archetype_name
      try:
        archetype = self._search_archetypes(montreal_custom_catalog, archetype_name)
      except KeyError:
        logging.error('Building %s has unknown energy system archetype for system name %s', building.name,
                      archetype_name)
        continue

      if archetype.name not in _generic_energy_systems:
        _generic_energy_systems = self._create_generic_systems_list(archetype, _generic_energy_systems)

    city.generic_energy_systems = _generic_energy_systems

    self._assign_energy_systems_to_buildings(city)

  @staticmethod
  def _search_archetypes(catalog, name):
    archetypes = catalog.entries('archetypes')
    for building_archetype in archetypes:
      if str(name) == str(building_archetype.name):
        return building_archetype
    raise KeyError('archetype not found')

  def _create_generic_systems_list(self, archetype, _generic_energy_systems):
    building_systems = []
    for archetype_system in archetype.systems:
      energy_system = EnergySystem()
      _hub_demand_types = []
      for demand_type in archetype_system.demand_types:
        _hub_demand_types.append(Dictionaries().montreal_demand_type_to_hub_energy_demand_type[demand_type])
      energy_system.name = archetype_system.name
      energy_system.demand_types = _hub_demand_types
      energy_system.generation_systems = self._create_generation_systems(archetype_system)
      energy_system.distribution_systems = self._create_distribution_systems(archetype_system)
      building_systems.append(energy_system)

    _generic_energy_systems[archetype.name] = building_systems

    return _generic_energy_systems

  @staticmethod
  def _create_generation_systems(archetype_system):
    _generation_systems = []
    for archetype_generation_system in archetype_system.generation_systems:
      if archetype_generation_system.system_type == 'photovoltaic':
        _generation_system = PvGenerationSystem()
        _type = archetype_generation_system.system_type
        _generation_system.system_type = Dictionaries().montreal_generation_system_to_hub_energy_generation_system[
          _type]
        _fuel_type = Dictionaries().montreal_custom_fuel_to_hub_fuel[archetype_generation_system.fuel_type]
        _generation_system.fuel_type = _fuel_type
        _generation_system.electricity_efficiency = archetype_generation_system.electricity_efficiency
        _generic_storage_system = None
        if archetype_generation_system.energy_storage_systems is not None:
          _generic_storage_system = ElectricalStorageSystem()
          _generic_storage_system.type_energy_stored = 'electrical'
        _generation_system.energy_storage_systems = [_generic_storage_system]
      else:
        _generation_system = NonPvGenerationSystem()
        _type = archetype_generation_system.system_type
        _generation_system.system_type = Dictionaries().montreal_generation_system_to_hub_energy_generation_system[
          _type]
        _fuel_type = Dictionaries().montreal_custom_fuel_to_hub_fuel[archetype_generation_system.fuel_type]
        _generation_system.fuel_type = _fuel_type
        _generation_system.source_types = archetype_generation_system.source_medium
        _generation_system.heat_efficiency = archetype_generation_system.heat_efficiency
        _generation_system.cooling_efficiency = archetype_generation_system.cooling_efficiency
        _generation_system.electricity_efficiency = archetype_generation_system.electricity_efficiency
        _generic_storage_system = None
        if archetype_generation_system.energy_storage_systems is not None:
          _storage_systems = []
          for storage_system in archetype_generation_system.energy_storage_systems:
            if storage_system.type_energy_stored == 'electrical':
              _generic_storage_system = ElectricalStorageSystem()
              _generic_storage_system.type_energy_stored = 'electrical'
            else:
              _generic_storage_system = ThermalStorageSystem()
              _generic_storage_system.type_energy_stored = storage_system.type_energy_stored
              _generic_storage_system.height = storage_system.height
              _generic_storage_system.layers = storage_system.layers
              _generic_storage_system.storage_medium = storage_system.storage_medium
            _storage_systems.append(_generic_storage_system)
          _generation_system.energy_storage_systems = _storage_systems
      _generation_systems.append(_generation_system)
    return _generation_systems

  @staticmethod
  def _create_distribution_systems(archetype_system):
    _distribution_systems = []
    for archetype_distribution_system in archetype_system.distribution_systems:
      _distribution_system = DistributionSystem()
      _distribution_system.type = archetype_distribution_system.type
      _distribution_system.distribution_consumption_fix_flow = \
        archetype_distribution_system.distribution_consumption_fix_flow
      _distribution_system.distribution_consumption_variable_flow = \
        archetype_distribution_system.distribution_consumption_variable_flow
      _distribution_system.heat_losses = archetype_distribution_system.heat_losses
      _generic_emission_system = None
      if archetype_distribution_system.emission_systems is not None:
        _emission_systems = []
        for emission_system in archetype_distribution_system.emission_systems:
          _generic_emission_system = EmissionSystem()
          _generic_emission_system.parasitic_energy_consumption = emission_system.parasitic_energy_consumption
          _emission_systems.append(_generic_emission_system)
        _distribution_system.emission_systems = _emission_systems
      _distribution_systems.append(_distribution_system)
    return _distribution_systems

  @staticmethod
  def _assign_energy_systems_to_buildings(city):
    for building in city.buildings:
      _building_energy_systems = []
      energy_systems_cluster_name = building.energy_systems_archetype_name
      if str(energy_systems_cluster_name) == 'nan':
        break
      _generic_building_energy_systems = city.generic_energy_systems[energy_systems_cluster_name]
      for _generic_building_energy_system in _generic_building_energy_systems:
        _building_energy_systems.append(copy.deepcopy(_generic_building_energy_system))

      building.energy_systems = _building_energy_systems



"""
Energy System catalog heat generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class NorthAmericaStorageSystemToHubEnergyStorage:
  """
  Montreal's system to hub energy generation system class
  """
  def __init__(self):
    self._dictionary = {
      'template Hot Water Storage Tank': cte.SENSIBLE,
                        }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Dictionaries module for hub function to Palma construction function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Cecilia Pérez cperez@irec.cat
"""

import hub.helpers.constants as cte


class HubFunctionToPalmaConstructionFunction:
  """
  Hub function to Palma construction function class
  """
  def __init__(self):
    self._dictionary = {
      cte.RESIDENTIAL: 'V',
      cte.SINGLE_FAMILY_HOUSE: 'Single-family building',
      cte.HIGH_RISE_APARTMENT: 'Large multifamily building',
      cte.MID_RISE_APARTMENT: 'Medium multifamily building',
      cte.MULTI_FAMILY_HOUSE: 'Small multifamily building'
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary


"""
DBFactory performs read related operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project CoderPeter Yefi peteryefi@gmail.com
"""
from typing import Dict

from hub.persistence.repositories.application import Application
from hub.persistence.repositories.city import City
from hub.persistence.repositories.city_object import CityObject
from hub.persistence.repositories.simulation_results import SimulationResults
from hub.persistence.repositories.user import User
from hub.persistence.repositories.user import UserRoles


class DBControl:
  """
  DBFactory class
  """

  def __init__(self, db_name, app_env, dotenv_path):
    self._city = City(db_name=db_name, dotenv_path=dotenv_path, app_env=app_env)
    self._application = Application(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._user = User(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._city_object = CityObject(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._simulation_results = SimulationResults(db_name=db_name, dotenv_path=dotenv_path, app_env=app_env)

  def application_info(self, application_uuid) -> Application:
    """
    Retrieve the application info for the given uuid from the database
    :param application_uuid: the uuid for the application
    :return: Application
    """
    return self._application.get_by_uuid(application_uuid)

  def user_info(self, name, password, application_id) -> User:
    """
    Retrieve the user info for the given name and password and application_id from the database
    :param name: the username
    :param password: the user password
    :param application_id: the application id
    :return: User
    """
    return self._user.get_by_name_application_id_and_password(name, password, application_id)

  def user_login(self, name, password, application_uuid) -> User:
    """
    Retrieve the user info from the database
    :param name: the username
    :param password: the user password
    :param application_uuid: the application uuid
    :return: User
    """
    return self._user.get_by_name_application_uuid_and_password(name, password, application_uuid)

  def cities_by_user_and_application(self, user_id, application_id) -> [City]:
    """
    Retrieve the cities belonging to the user and the application from the database
    :param user_id: User id
    :param application_id: Application id
    :return: [City]
    """
    return self._city.get_by_user_id_and_application_id(user_id, application_id)

  def building(self, name, user_id, application_id, scenario) -> CityObject:
    """
    Retrieve the building from the database
    :param name: Building name
    :param user_id: User id
    :param application_id: Application id
    :param scenario: Scenario
    :
    """
    cities = self._city.get_by_user_id_application_id_and_scenario(user_id, application_id, scenario)
    c = [c[0].id for c in cities]
    result = self._city_object.building_in_cities_info(name, c)
    if result is not None:
      return result
    return None

  def building_info(self, name, city_id) -> CityObject:
    """
    Retrieve the building info from the database
    :param name: Building name
    :param city_id: City ID
    :return: CityObject
    """
    return self._city_object.get_by_name_or_alias_and_city(name, city_id)

  def building_info_in_cities(self, name, cities) -> CityObject:
    """
    Retrieve the building info from the database
    :param name: Building name
    :param cities: [City ID]
    :return: CityObject
    """
    return self._city_object.get_by_name_or_alias_in_cities(name, cities)

  def buildings_info(self, user_id, application_id, names_or_aliases) -> [CityObject]:
    """
    Retrieve the buildings info from the database
    :param user_id: User ID
    :param application_id: Application ID
    :param names_or_aliases: A list of names or alias for the buildings
    :return: [CityObject]
    """
    results = self._city_object.get_by_name_or_alias_for_user_app(user_id, application_id, names_or_aliases)
    if results is None:
      return []
    return results

  def results(self, user_id, application_id, request_values, result_names=None) -> Dict:
    """
    Retrieve the simulation results for the given cities from the database
    :param user_id: the user id owning the results
    :param application_id: the application id owning the results
    :param request_values: dictionary containing the scenario and building names to grab the results
    :param result_names: if given, filter the results to the selected names
    """
    if result_names is None:
      result_names = []
    results = {}
    for scenario in request_values['scenarios']:
      for scenario_name in scenario.keys():
        result_sets = self._city.get_by_user_id_application_id_and_scenario(
          user_id,
          application_id,
          scenario_name
        )
        if result_sets is None:
          continue
        results[scenario_name] = []
        city_ids = [r[0].id for r in result_sets]
        for building_name in scenario[scenario_name]:
          _building = self._city_object.get_by_name_or_alias_in_cities(building_name, city_ids)
          if _building is None:
            continue
          city_object_id = _building.id
          _ = self._simulation_results.get_simulation_results_by_city_object_id_and_names(
            city_object_id,
            result_names)

          for value in _:
            values = value.values
            values["building"] = building_name
            results[scenario_name].append(values)
    return results

  def persist_city(self, city: City, pickle_path, scenario, application_id: int, user_id: int):
    """
    Creates a city into the database
    :param city: City to be stored
    :param pickle_path: Path to save the pickle file
    :param scenario: Simulation scenario name
    :param application_id: Application id owning this city
    :param user_id: User who create the city
    return identity_id
    """
    return self._city.insert(city, pickle_path, scenario,  application_id, user_id)

  def update_city(self, city_id, city):
    """
   Update an existing city in the database
   :param city_id: the id of the city to update
   :param city: the updated city object
    """
    return self._city.update(city_id, city)

  def persist_application(self, name: str, description: str, application_uuid: str):
    """
    Creates information for an application in the database
   :param name: name of application
   :param description: the description of the application
   :param application_uuid: the uuid of the application to be created
    """
    return self._application.insert(name, description, application_uuid)

  def update_application(self, name: str, description: str, application_uuid: str):
    """
    Update the application information stored in the database
   :param name: name of application
   :param description: the description of the application
   :param application_uuid: the uuid of the application to be created
    """
    return self._application.update(application_uuid, name, description)

  def add_simulation_results(self, name, values, city_id=None, city_object_id=None):
    """
    Add simulation results to the city or to the city_object to the database
    :param name: simulation and simulation engine name
    :param values: simulation values in json format
    :param city_id: city id or None
    :param city_object_id: city object id or None
    """
    return self._simulation_results.insert(name, values, city_id, city_object_id)

  def create_user(self, name: str, application_id: int, password: str, role: UserRoles):
    """
    Creates a new user in the database
    :param name: the name of the user
    :param application_id: the application id of the user
    :param password: the password of the user
    :param role: the role of the user
    """
    return self._user.insert(name, password, role, application_id)

  def update_user(self, user_id: int, name: str, password: str, role: UserRoles):
    """
    Updates a user in the database
    :param user_id: the id of the user
    :param name: the name of the user
    :param password: the password of the user
    :param role: the role of the user
    """
    return self._user.update(user_id, name, password, role)

  def get_by_name_and_application(self, name: str, application: int):
    """
    Retrieve a single user from the database
    :param name: username
    :param application: application accessing hub
    """
    return self._user.get_by_name_and_application(name, application)

  def delete_user(self, user_id):
    """
    Delete a single user from the database
    :param user_id: the id of the user to delete
    """
    self._user.delete(user_id)

  def delete_city(self, city_id):
    """
    Deletes a single city from the database
    :param city_id: the id of the city to get
    """
    self._city.delete(city_id)

  def delete_results_by_name(self, name, city_id=None, city_object_id=None):
    """
    Deletes city object simulation results from the database
    :param name: simulation name
    :param city_id: if given, delete delete the results for the city with id city_id
    :param city_object_id: if given, delete delete the results for the city object with id city_object_id
    """
    self._simulation_results.delete(name, city_id=city_id, city_object_id=city_object_id)

  def delete_application(self, application_uuid):
    """
    Deletes a single application from the database
    :param application_uuid: the id of the application to get
    """
    self._application.delete(application_uuid)

  def get_city(self, city_id):
    return self._city.get_by_id(city_id)
